import warnings
from typing import Union, Iterable, Tuple, Optional, Any, Literal
from datetime import date as dt_date, datetime, timedelta

import pandas as pd
from dateutil.tz import gettz

import onetick.py as otp
from onetick.py import configuration, utils
from onetick.py.compatibility import is_native_plus_zstd_supported
from onetick.py.core import db_constants
from onetick.py.otq import otq


def _datetime2date(dt: Union[dt_date, datetime]) -> dt_date:
    """ Convert datetime and date explicitly into the datetime.date """
    return dt_date(dt.year, dt.month, dt.day)


class DB:

    """
    An object of available databases that the :py:func:`otp.databases() <onetick.py.databases>` function returns.
    It helps to make initial analysis on the database level: available tick types,
    dates with data, symbols, tick schema, etc.
    """

    def __init__(self, name, context=utils.default):
        self.name = name
        if context is utils.default or context is None:
            self.context = otp.config.context
        else:
            self.context = context
        self._locator_date_ranges = None

    def __eq__(self, obj):
        return str(self) == str(obj)

    def __lt__(self, obj):
        return str(self) < str(obj)

    def __str__(self):
        return self.name

    def access_info(self, deep_scan=False, username=None) -> Union[pd.DataFrame, dict]:
        """
        Get access info for this database and ``username``.

        All dates are returned in GMT timezone.

        Parameters
        ----------
        deep_scan:
            If False (default) then the access fields are returned from the configuration of the database
            (basically the same fields as specified in the locator) and the dictionary is returned.
            If True then access fields are returned for each available remote host and time interval
            and the :pandas:`pandas.DataFrame` object is returned.
        username:
            Can be used to specify the user for which the query will be executed.
            By default the query is executed for the current user.

        See also
        --------
        **ACCESS_INFO** OneTick event processor

        Examples
        --------

        By default access fields from the basic configuration of the database are returned:

        >>> some_db = otp.databases()['SOME_DB']
        >>> some_db.access_info()  # doctest: +SKIP
        {'DB_NAME': 'SOME_DB',
         'READ_ACCESS': 1,
         'WRITE_ACCESS': 1,
         'MIN_AGE_SET': 0,
         'MIN_AGE_MSEC': 0,
         'MAX_AGE_SET': 0,
         'MAX_AGE_MSEC': 0,
         'MIN_START_DATE_SET': 0,
         'MIN_START_DATE_MSEC': Timestamp('1970-01-01 00:00:00'),
         'MAX_END_DATE_SET': 0,
         'MAX_END_DATE_MSEC': Timestamp('1970-01-01 00:00:00'),
         'MIN_AGE_DB_DAYS': 0,
         'MIN_AGE_DB_DAYS_SET': 0,
         'MAX_AGE_DB_DAYS': 0,
         'MAX_AGE_DB_DAYS_SET': 0,
         'CEP_ACCESS': 1,
         'DESTROY_ACCESS': 0,
         'MEMDB_ACCESS': 1}

        Set parameter ``deep_scan`` to True to return access fields from each available host and time interval:

        >>> some_db.access_info(deep_scan=True)  # doctest: +SKIP
           DB_NAME  READ_ACCESS  WRITE_ACCESS  MIN_AGE_SET  MIN_AGE_MSEC  MAX_AGE_SET  MAX_AGE_MSEC\
              MIN_START_DATE_SET MIN_START_DATE_MSEC  MAX_END_DATE_SET MAX_END_DATE_MSEC  MIN_AGE_DB_DAYS\
              MIN_AGE_DB_DAYS_SET  MAX_AGE_DB_DAYS  MAX_AGE_DB_DAYS_SET  CEP_ACCESS  DESTROY_ACCESS  MEMDB_ACCESS\
                SERVER_ADDRESS INTERVAL_START INTERVAL_END
        0  SOME_DB            1             1            0             0            0             0\
                               0          1970-01-01                 0        1970-01-01                0\
                                0                0                    0           1               0             1\
                           ...     2002-12-30   2100-01-01
        """
        # get parent name for derived databases, only parent databases will be listed by AccessInfo
        name, _, _ = self.name.partition('//')
        node = (
            otq.AccessInfo(info_type='DATABASES', show_for_all_users=False, deep_scan=deep_scan)
            >> otq.WhereClause(where=f'DB_NAME = "{name}"')
        )
        graph = otq.GraphQuery(node)
        df = otp.run(graph,
                     symbols='LOCAL::',
                     # start and end times don't matter
                     start=db_constants.DEFAULT_START_DATE,
                     end=db_constants.DEFAULT_END_DATE,
                     # and timezone is GMT, because timestamp parameters in ACL are in GMT
                     timezone='GMT',
                     username=username,
                     context=self.context)
        if not df.empty:
            df = df.drop(columns='Time')
        if deep_scan:
            return df
        return dict(df.iloc[0] if not df.empty else {})

    def show_config(self, config_type: Literal['locator_entry', 'db_time_intervals'] = 'locator_entry') -> dict:
        """
        Shows the specified configuration for a database.

        Parameters
        ----------
        config_type: str
            If **'locator_entry'** is specified, a string representing db's locator entry along with VDB_FLAG
            (this flag equals 1 when the database is virtual and 0 otherwise) will be returned.

            If **'db_time_intervals'** is specified,
            then time intervals configured in the locator file will be propagated
            including additional information, such as
            LOCATION, ARCHIVE_DURATION, DAY_BOUNDARY_TZ, DAY_BOUNDARY_OFFSET, ALTERNATIVE_LOCATIONS, etc.

        See also
        --------
        **DB/SHOW_CONFIG** OneTick event processor

        Examples
        --------
        .. testcode::
           :skipif: not is_native_plus_zstd_supported()

           some_db = otp.databases()['SOME_DB']
           print(some_db.show_config()['LOCATOR_STRING'])

        .. testoutput::
           :options: +ELLIPSIS

           <DB ARCHIVE_COMPRESSION_TYPE="NATIVE_PLUS_ZSTD" ID="SOME_DB" SYMBOLOGY="BZX" TICK_TIMESTAMP_TYPE="NANOS" >
           <LOCATIONS >
               <LOCATION ACCESS_METHOD="file" DAY_BOUNDARY_TZ="EST5EDT"
                         END_TIME="21000101000000" LOCATION="..." START_TIME="20021230000000" />
           </LOCATIONS>
           <RAW_DATA />
           </DB>

        >>> some_db = otp.databases()['SOME_DB']
        >>> some_db.show_config(config_type='db_time_intervals')  # doctest: +ELLIPSIS
        {'START_DATE': 1041206400000, 'END_DATE': 4102444800000,
         'GROWABLE_ARCHIVE_FLAG': 0, 'ARCHIVE_DURATION': 0,
         'LOCATION': '...', 'DAY_BOUNDARY_TZ': 'EST5EDT', 'DAY_BOUNDARY_OFFSET': 0, 'ALTERNATIVE_LOCATIONS': ''}
        """
        node = otq.DbShowConfig(db_name=self.name, config_type=config_type.upper())
        graph = otq.GraphQuery(node)
        df = otp.run(graph,
                     symbols='LOCAL::',
                     # start and end times don't matter
                     start=db_constants.DEFAULT_START_DATE,
                     end=db_constants.DEFAULT_END_DATE,
                     # and timezone is GMT, because timestamp parameters in ACL are in GMT
                     timezone='GMT',
                     context=self.context)
        if df.empty:
            raise ValueError(f"Can't get config for database '{self.name}'")
        df = df.drop(columns='Time')
        return dict(df.iloc[0])

    @property
    def min_acl_start_date(self) -> Optional[dt_date]:
        """
        Minimum start date set in ACL for current user.
        Returns None if not set.
        """
        access_info = self.access_info()
        if not access_info:
            return None
        if access_info['MIN_START_DATE_SET'] == 0:
            return None
        return _datetime2date(access_info['MIN_START_DATE_MSEC'])

    @property
    def max_acl_end_date(self) -> Optional[dt_date]:
        """
        Maximum end date set in ACL for current user.
        Returns None if not set.
        """
        access_info = self.access_info()
        if not access_info:
            return None
        if access_info['MAX_END_DATE_SET'] == 0:
            return None
        return _datetime2date(access_info['MAX_END_DATE_MSEC'])

    def _fit_time_interval_in_acl(self, start, end, timezone='GMT') -> Tuple[datetime, datetime]:
        """
        Returns the part of time interval between ``start`` and ``end`` that fits ACL start/end time rules.
        ``start`` and ``end`` objects are considered to be timezone-naive and will be localized in ``timezone``.

        If it's not possible to find such interval, raises ValueError.
        """
        # convert to GMT, because ACL timestamps are in GMT
        start = otp.dt(utils.convert_timezone(start, timezone, 'GMT'))
        end = otp.dt(utils.convert_timezone(end, timezone, 'GMT'))

        if self.min_acl_start_date is not None:
            if end < otp.dt(self.min_acl_start_date):
                # fully not intersecting intervals
                raise ValueError(f'Date {start.date()} {timezone} violates ACL rules for the database {self.name}:'
                                 f' minimum start time is {otp.dt(self.min_acl_start_date)} GMT.')
            # partly intersecting intervals, choose the part not violating ACL
            start = max(start, otp.dt(self.min_acl_start_date))

        if self.max_acl_end_date is not None:
            if start >= otp.dt(self.max_acl_end_date):
                # fully not intersecting intervals
                raise ValueError(f'Date {start.date()} {timezone} violates ACL rules for the database {self.name}:'
                                 f' maximum (exclusive) end time is {otp.dt(self.max_acl_end_date)} GMT.')
            # partly intersecting intervals, choose the part not violating ACL
            end = min(end, otp.dt(self.max_acl_end_date))

        # convert back to timezone
        start = utils.convert_timezone(start, 'GMT', timezone)
        end = utils.convert_timezone(end, 'GMT', timezone)
        return start, end

    def _fit_date_in_acl(self, date, timezone='GMT') -> Tuple[datetime, datetime]:
        """
        Returns the part of ``date`` time interval that fits ACL start/end time rules.
        ``date`` object is considered to be timezone-naive and will be localized in ``timezone``.

        If it's not possible to find such interval, raises ValueError.
        """
        date = _datetime2date(date)
        start = otp.dt(date)
        end = start + otp.Day(1)
        return self._fit_time_interval_in_acl(start, end, timezone)

    def _set_intervals(self):
        """
        Finds all date ranges from locators.
        These intervals are required to find all possible dates with data.
        It is only possible by querying the DB_SHOW_LOADED_TIME_RANGE
        against the largest possible query date range.
        """

        if self._locator_date_ranges is None:
            graph = otq.GraphQuery(otq.DbShowConfiguredTimeRanges(db_name=self.name).tick_type("ANY")
                                   >> otq.Table(fields='long START_DATE, long END_DATE'))

            result = otp.run(graph,
                             symbols=f'{self.name}::',
                             # start and end times don't matter for this query, use some constants
                             start=db_constants.DEFAULT_START_DATE,
                             end=db_constants.DEFAULT_END_DATE,
                             # GMT, because start/end timestamp in locator are in GMT
                             timezone='GMT',
                             context=self.context)

            date_ranges = []

            tz_gmt = gettz('GMT')
            for inx in range(len(result)):
                start_date = result['START_DATE'][inx]
                # On Windows datetime.fromtimestamp throws an OSError for negative values
                start_date = max(start_date, 0)
                start = datetime.fromtimestamp(start_date / 1000, tz=tz_gmt)
                start = start.replace(tzinfo=None)
                try:
                    end = datetime.fromtimestamp(result['END_DATE'][inx] / 1000, tz=tz_gmt)
                except (ValueError, OSError):
                    # this may happen if value exceeds 9999-12-31 23:59:59.999999
                    end = datetime.max
                end = end.replace(tzinfo=None)

                date_ranges.append((start, end))

            # merge ranges if necessary to reduce number of queries
            # for `dates` property then
            self._locator_date_ranges = []
            start, end = None, None

            for t_start, t_end in date_ranges:
                if start is None:
                    start = t_start
                if end is None:
                    end = t_end
                else:
                    if t_start == end:
                        end = t_end
                    else:
                        self._locator_date_ranges.append((start, end))
                        start, end = t_start, t_end

            if start and end:
                self._locator_date_ranges.append((start, end))

    def _show_loaded_time_ranges(self, start, end, only_last=False, prefer_speed_over_accuracy=False):
        kwargs = {}
        if prefer_speed_over_accuracy:
            kwargs['prefer_speed_over_accuracy'] = prefer_speed_over_accuracy

        eps = otq.DbShowLoadedTimeRanges(use_cache=True, **kwargs).tick_type('ANY')
        eps = eps >> otq.WhereClause(where='NUM_LOADED_PARTITIONS > 0')
        if only_last:
            eps = eps >> otq.LastTick()

        graph = otq.GraphQuery(eps)
        result = otp.run(graph,
                         symbols=f'{self.name}::',
                         start=start,
                         end=end,
                         # GMT works properly for locators with gap
                         timezone='GMT',
                         context=self.context)

        dates = []
        # every record contains consequent intervals of data on disk
        for inx in range(len(result)):
            start = datetime.strptime(str(result['START_DATE'][inx]), '%Y%m%d')
            end = datetime.strptime(str(result['END_DATE'][inx]), '%Y%m%d')
            if only_last:
                return [_datetime2date(end)]
            while start <= end:
                dates.append(_datetime2date(start))
                start += timedelta(days=1)

        return dates

    def __split_loaded_time_ranges(self, locator_start, locator_end, only_last):
        # locator date range can be very big, so splitting it in smaller parts
        # (because _show_loaded_time_ranges() can be very slow for big time ranges)
        # it is especially useful when we only need the last date
        dates = []
        start = end = locator_end
        delta = 1 if only_last else 365
        while locator_start < start:
            start = end - timedelta(days=delta)
            start = max(locator_start, start)
            loaded_dates = self._show_loaded_time_ranges(start, end, only_last=only_last)
            if only_last and loaded_dates:
                return [loaded_dates[-1]]
            dates = loaded_dates + dates
            end = start
            # if we are not getting data, then increasing time range to find it faster
            if not loaded_dates:
                delta *= 2
        return dates

    def __get_dates(self, only_last=False, respect_acl=False, check_index_file=utils.adaptive):
        """ Returns list of dates in GMT timezone with data """
        self._set_intervals()

        dates = []
        today = dt_date.today()
        today = datetime(today.year, today.month, today.day)
        # searching in reversed order in case we need only_last date
        for locator_start, locator_end in reversed(self._locator_date_ranges):
            # future is not loaded yet
            if locator_start > today:
                continue
            locator_end = min(locator_end, today)

            if respect_acl:
                try:
                    locator_start, locator_end = self._fit_time_interval_in_acl(locator_start, locator_end)
                except ValueError:
                    # fully not intersecting intervals, trying next locator date range
                    continue

            if check_index_file is utils.adaptive or check_index_file is None:
                prefer_speed_over_accuracy = True
            else:
                prefer_speed_over_accuracy = not check_index_file
            try:
                loaded_dates = self._show_loaded_time_ranges(locator_start, locator_end,
                                                             only_last=only_last,
                                                             prefer_speed_over_accuracy=prefer_speed_over_accuracy)
            except Exception as e:
                # parameter prefer_speed_over_accuracy is not supported on all OneTick versions and servers
                if check_index_file is not utils.adaptive:
                    raise ValueError(
                        "Parameter 'check_index_file' is not supported by the API or OneTick server"
                    ) from e
                # in this case we fall back to splitting the locator range into smaller parts to increase speed
                loaded_dates = self.__split_loaded_time_ranges(locator_start, locator_end, only_last)

            if only_last and loaded_dates:
                return loaded_dates[-1]
            dates = loaded_dates + dates

        if only_last and len(dates) == 0:
            return None  # no data on disk

        return dates

    def dates(self, respect_acl=False, check_index_file=utils.adaptive):
        """
        Returns list of dates in GMT timezone for which data is available.

        Parameters
        ----------
        respect_acl: bool
            If True then only the dates that current user has access to will be returned
        check_index_file: bool
            If True, then file *index* will be searched for to determine if a database is loaded for a date.
            This check may be expensive, in terms of time it takes,
            when the file resides on NFS or on object storage, such as S3.
            If this parameter is set to False, then only the database directory for a date will be searched.
            This will increase performance, but may also return the days that are configured
            but where there is actually no data.
            By default this option is set to False if it is supported by API and the server,
            otherwise it is set to True.

        Returns
        -------
        ``datetime.date`` or ``None``
            Returns ``None`` when there is no data in the database

        Examples
        --------
        >>> some_db = otp.databases()['SOME_DB']
        >>> some_db.dates()
        [datetime.date(2003, 12, 1)]
        """
        return self.__get_dates(respect_acl=respect_acl, check_index_file=check_index_file)

    def last_not_empty_date(self, last_date, days_back, timezone=None, tick_type=None):
        """
        Find first day that has data
        starting from ``last_date`` and going ``days_back`` number of days back.
        """
        min_locator_date = self.min_locator_date()
        for i in range(days_back + 1):
            date = _datetime2date(last_date - timedelta(days=i))
            if date < min_locator_date:
                break
            try:
                tick_types = self.tick_types(date, timezone=timezone)
            except ValueError:
                # acl date violation
                break
            if tick_type is None and tick_types:
                return date
            if tick_type is not None and tick_type in tick_types:
                return date
        return None

    @property
    def last_date(self):
        """
        The latest date on which db has data and the current user has access to.

        Returns
        -------
        ``datetime.date`` or ``None``
            Returns ``None`` when there is no data in the database

        Examples
        --------
        >>> some_db = otp.databases()['SOME_DB']
        >>> some_db.last_date
        datetime.date(2003, 12, 1)
        """
        return self.get_last_date()

    def get_last_date(self, tick_type=None, timezone=None, show_warnings=True, check_index_file=utils.adaptive):
        last_date = self.__get_dates(only_last=True, respect_acl=True, check_index_file=check_index_file)
        if last_date is None:
            return None
        # It might happen that database loading processes is configured
        # to work over weekends and holidays and therefore
        # there are days that are configured but have no data, tick types and schema.
        # We want to find the closest not empty day because
        # we want to expose the most actual schema to end user.
        # For example, this is a case of OneTick Cloud US_COMP database.
        # We only scan 5 previous days to cover weekends + possible conjuncted holidays.
        # According to the official NYSE calendar there are no more than 5 closed days.
        date = self.last_not_empty_date(last_date, days_back=5, tick_type=tick_type, timezone=timezone)
        if date is None:
            if show_warnings:
                warnings.warn(
                    "Can't find not empty day for the last 5 days, using last configured day. "
                    "Try to use .last_not_empty_date() function to find older not empty days."
                )
            return last_date
        return date

    def tick_types(self, date=None, timezone=None) -> list[str]:
        """
        Returns list of tick types for the ``date``.

        Parameters
        ----------
        date: :class:`otp.dt <onetick.py.datetime>`, :py:class:`datetime.datetime`, optional
            Date for the tick types look up. ``None`` means the :attr:`last_date`
        timezone: str, optional
            Timezone for the look up. ``None`` means the default timezone.

        Returns
        -------
        list
            List with string values of available tick types.

        Examples
        --------
        >>> us_comp_db = otp.databases()['US_COMP']
        >>> us_comp_db.tick_types(date=otp.dt(2022, 3, 1))
        ['QTE', 'TRD']
        """
        date = self.last_date if date is None else date
        if timezone is None:
            timezone = configuration.config.tz
        time_params: dict[str, Any] = {}

        if date is not None:
            time_params['start'], time_params['end'] = self._fit_date_in_acl(date, timezone=timezone)

        # PY-458: don't use cache, it can return different result in some cases
        result = otp.run(otq.DbShowTickTypes(use_cache=False,
                                             show_schema=False,
                                             include_memdb=True),
                         symbols=f'{self.name}::',
                         **time_params,
                         timezone=timezone,
                         context=self.context)

        if len(result) == 0:
            return []

        return result['TICK_TYPE_NAME'].tolist()

    def min_locator_date(self):
        self._set_intervals()
        min_date = min(obj[0] for obj in self._locator_date_ranges)
        return _datetime2date(min_date)

    def schema(self, date=None, tick_type=None, timezone=None, check_index_file=utils.adaptive) -> dict[str, type]:
        """
        Gets the schema of the database.

        Parameters
        ----------
        date: :class:`otp.dt <onetick.py.datetime>`, :py:class:`datetime.datetime`, optional
            Date for the schema. ``None`` means the :attr:`last_date`
        tick_type: str, optional
            Specifies a tick type for schema. ``None`` means use the one available
            tick type, if there are multiple tick types then it raises the ``Exception``.
            It uses the :meth:`tick_types` method.
        timezone: str, optional
            Allows to specify a timezone for searching tick types.
        check_index_file: bool
            If True, then file *index* will be searched for to determine if a database is loaded for a date.
            This check may be expensive, in terms of time it takes,
            when the file resides on NFS or on object storage, such as S3.
            If this parameter is set to False, then only the database directory for a date will be searched.
            This will increase performance, but may also return the days that are configured
            but where there is actually no data.
            By default this option is set to False if it is supported by API and the server,
            otherwise it is set to True.

        Returns
        -------
        dict
            Dict where keys are field names and values are ``onetick.py`` :ref:`types <schema concept>`.
            It's compatible with the :attr:`onetick.py.Source.schema` methods.

        Examples
        --------
        >>> us_comp_db = otp.databases()['US_COMP']
        >>> us_comp_db.schema(tick_type='TRD', date=otp.dt(2022, 3, 1))
        {'PRICE': <class 'float'>, 'SIZE': <class 'int'>}
        """
        orig_date = date

        if date is None:
            date = self.get_last_date(tick_type=tick_type, timezone=timezone, check_index_file=check_index_file)
        if timezone is None:
            timezone = configuration.config.tz
        if tick_type is None:
            tick_types = self.tick_types(date=date, timezone=timezone)
            if len(tick_types) == 0:
                raise ValueError("No tick types has found and specified")
            if len(tick_types) > 1:
                raise ValueError("Database has multiple tick types, please specify using the `tick_type` parameter")

            tick_type = tick_types[0]

        if date is None:
            # it might happen when a database has no data on disks
            return {}

        # Convert explicitly into the datetime.date, because min_date and date
        # could be date or datetime types, and datetime is not comparable with datetime.date
        date = _datetime2date(date)

        start, end = self._fit_date_in_acl(date, timezone=timezone)

        # TODO: refactor into global method, use in tick_types()
        def get_schema(use_cache: bool = True):
            return otp.run(otq.DbShowTickTypes(use_cache=use_cache,
                                               show_schema=True,
                                               include_memdb=True)
                           >> otq.WhereClause(where=f'TICK_TYPE_NAME="{tick_type}"'),
                           symbols=f'{self.name}::',
                           start=start,
                           end=end,
                           timezone=timezone,
                           context=self.context)

        result = get_schema(use_cache=True)
        if result.empty:
            # in case cache settings in database are bad (e.g. BEXRTS-1220)
            result = get_schema(use_cache=False)

        fields: Iterable
        if len(result):
            # filter schema by date
            date_to_filter = None
            if orig_date:
                # if date is passed as a parameter -- then use it
                date_to_filter = date
            else:
                # otherwise use the closest date
                date_to_filter = result['Time'].max()

            result = result[(result['Time'] >= pd.Timestamp(date_to_filter))]

            fields = zip(result['FIELD_NAME'].tolist(),
                         result['FIELD_TYPE_NAME'].tolist(),
                         result['FIELD_SIZE'].tolist())
        else:
            fields = []

        schema = {}

        for fname, ftype, fsize in fields:
            dtype: type

            if 'UINT32' in ftype:
                dtype = otp.uint
            elif 'UINT64' in ftype:
                dtype = otp.ulong
            elif 'INT32' in ftype:
                dtype = otp.int
            elif 'INT64' in ftype:
                # otp.long can be used too, but we use int for backward compatibility
                dtype = int
            elif 'INT8' in ftype:
                dtype = otp.byte
            elif 'INT16' in ftype:
                dtype = otp.short
            elif 'INT' in ftype:
                dtype = int
            elif 'MSEC' in ftype:
                dtype = otp.msectime
            elif 'NSEC' in ftype:
                dtype = otp.nsectime
            elif 'DOUBLE' in ftype or 'FLOAT' in ftype:
                dtype = float
            elif 'DECIMAL' in ftype:
                dtype = otp.decimal
            elif 'VARSTRING' in ftype:
                dtype = otp.varstring
            elif 'STRING' in ftype:
                if fsize == 64:
                    dtype = str
                else:
                    dtype = otp.string[fsize]
            else:
                warnings.warn(
                    f"Unsupported field type '{ftype}' for field '{fname}'. "
                    "Note that this field will be ignored "
                    "and will not be added to the python schema, "
                    "but will still remain in the OneTick schema."
                )
                continue

            schema[fname] = dtype

        return schema

    def symbols(self, date=None, timezone=None, tick_type=None, pattern='.*') -> list[str]:
        """
        Finds a list of available symbols in the database

        Parameters
        ----------
        date: :class:`otp.dt <onetick.py.datetime>`, :py:class:`datetime.datetime`, optional
            Date for the symbols look up. ``None`` means the :attr:`last_date`
        tick_type: str, optional
            Tick type for symbols. ``None`` means union across all tick types.
        timezone: str, optional
            Timezone for the lookup. ``None`` means the default timezone.
        pattern: str
            Regular expression to select symbols.

        Examples
        --------
        >>> us_comp_db = otp.databases()['US_COMP']
        >>> us_comp_db.symbols(date=otp.dt(2022, 3, 1), tick_type='TRD', pattern='^AAP.*')
        ['AAP', 'AAPL']
        """
        if date is None:
            date = self.last_date
        if timezone is None:
            timezone = configuration.config.tz
        if tick_type is None:
            tick_type = ''

        eps = otq.FindDbSymbols(pattern='%', tick_type_field=tick_type) \
            >> otq.AddField(field='varstring SYMBOL', value='regex_replace(SYMBOL_NAME, ".*::", "")') \
            >> otq.WhereClause(where=f'regex_match(SYMBOL, "{pattern}")') \
            >> otq.Table(fields='SYMBOL')

        result = otp.run(eps,
                         symbols=f'{self.name}::',
                         start=date,
                         end=date + timedelta(days=1),
                         timezone=timezone,
                         context=self.context)

        if len(result) == 0:
            return []

        return result['SYMBOL'].tolist()

    def show_archive_stats(
        self,
        start=utils.adaptive,
        end=utils.adaptive,
        date=None,
        timezone='GMT',
    ) -> pd.DataFrame:
        """
        This method shows various stats about the queried symbol,
        as well as an archive as a whole for each day within the queried interval.

        Accelerator databases are not supported.
        Memory databases will be ignored even within their life hours.

        Archive stats returned:

            * COMPRESSION_TYPE - archive compression type.
              In older archives native compression flag is not stored,
              so for example for gzip compression this field may say "GZIP or NATIVE_PLUS_GZIP".
              The meta_data_upgrader.exe tool can be used to determine and inject that information in such cases
              in order to get a more precise result in this field.
            * TIME_RANGE_VALIDITY - whether lowest and highest loaded timestamps (see below) are known.
              Like native compression flag, this information is missing in older archives
              and can be added using meta_data_upgrader.exe tool.
            * LOWEST_LOADED_DATETIME - the lowest loaded timestamp for the queried interval (across all symbols)
            * HIGHEST_LOADED_DATETIME - the highest loaded timestamp for the queried interval (across all symbols)
            * TOTAL_TICKS - the number of ticks for the queried interval (across all symbols).
              Also missing in older archives and can be added using meta_data_upgrader.exe.
              If not available, -1 will be returned.
            * SYMBOL_DATA_SIZE - the size of the symbol in archive in bytes.
              This information is also missing in older archives, however the other options, it cannot later be added.
              In such cases -1 will be returned.
            * TOTAL_SYMBOLS - the number of symbols for the queried interval
            * TOTAL_SIZE - archive size in bytes for the queried interval
              (including the garbage potentially accumulated during appends).

        Note
        ----
        Fields **LOWEST_LOADED_DATETIME** and **HIGHEST_LOADED_DATETIME** are returned in GMT timezone,
        so the default value of parameter ``timezone`` is GMT too.

        See also
        --------
        **SHOW_ARCHIVE_STATS** OneTick event processor

        Examples
        --------

        Show stats for a particular date for a database SOME_DB:

        .. testcode::
           :skipif: not is_native_plus_zstd_supported()

           db = otp.databases()['SOME_DB']
           stats = db.show_archive_stats(date=otp.dt(2003, 12, 1))
           print(stats)

        .. testoutput::
           :options: +ELLIPSIS

                            Time  COMPRESSION_TYPE TIME_RANGE_VALIDITY LOWEST_LOADED_DATETIME HIGHEST_LOADED_DATETIME...
           0 2003-12-01 05:00:00  NATIVE_PLUS_ZSTD               VALID    2003-12-01 05:00:00 2003-12-01 05:00:00.002...
        """
        node = otq.ShowArchiveStats()
        graph = otq.GraphQuery(node)
        df = otp.run(graph,
                     symbols=f'{self.name}::',
                     start=start,
                     end=end,
                     date=date,
                     timezone=timezone,
                     context=self.context)
        return df

    def ref_data(
        self,
        ref_data_type: str,
        symbol_date=None,
        start=utils.adaptive,
        end=utils.adaptive,
        date=None,
        timezone='GMT',
        symbol: str = '',
    ) -> pd.DataFrame:
        """
        Shows reference data for the specified security and reference data type.

        It can be used to view corporation actions,
        symbol name changes,
        primary exchange info and symbology mapping for a securities,
        as well as the list of symbologies,
        names of custom adjustment types for corporate actions present in a reference database
        as well as names of continuous contracts in database symbology.

        Parameters
        ----------
        ref_data_type: str
            Type of reference data to be queried. Possible values are:

                * corp_actions
                * symbol_name_history
                * primary_exchange
                * symbol_calendar
                * symbol_currency
                * symbology_mapping
                * symbology_list
                * custom_adjustment_type_list
                * all_calendars
                * all_continuous_contract_names
        symbol_date:
            This parameter must be specified for some reference data types to be queried.
        symbol:
            Symbol name for the query (may be useful for some ``ref_data_type``).

        See also
        --------
        **REF_DATA** OneTick event processor

        Examples
        --------

        Show calendars for a database TRAIN_A_PRL_TRD in the given range:

        >>> db = otp.databases()['TRAIN_A_PRL_TRD']  # doctest: +SKIP
        >>> db.ref_data('all_calendars',  # doctest: +SKIP
        ...             start=otp.dt(2018, 2, 1),
        ...             end=otp.dt(2018, 2, 9),
        ...             symbol_date=otp.dt(2018, 2, 1))
                         Time        END_DATETIME CALENDAR_NAME SESSION_NAME SESSION_FLAGS DAY_PATTERN  START_HHMMSS\
          END_HHMMSS TIMEZONE  PRIORITY DESCRIPTION
        0 2018-02-01 00:00:00 2018-02-06 23:59:59          FRED      Regular             R   0.0.12345         93000\
              160000  EST5EDT         0
        1 2018-02-06 23:59:59 2018-02-07 23:59:59          FRED      Holiday             H   0.0.12345         93000\
              160000  EST5EDT         1
        2 2018-02-07 23:59:59 2050-12-31 23:59:59          FRED      Regular             F   0.0.12345         93000\
              160000  EST5EDT         0

        Set symbol name with ``symbol`` parameter:

        >>> db = otp.databases()['US_COMP_SAMPLE']  # doctest: +SKIP
        >>> db.ref_data(ref_data_type='corp_actions',  # doctest: +SKIP
        ...             start=otp.dt(2025, 1, 2),
        ...             end=otp.dt(2025, 7, 2),
        ...             symbol_date=otp.dt(2025, 7, 1),
        ...             symbol='WMT',
        ...             timezone='America/New_York')
                Time  MULTIPLICATIVE_ADJUSTMENT  ADDITIVE_ADJUSTMENT ADJUSTMENT_TYPE
        0 2025-03-21                   1.000000                0.235   CASH_DIVIDEND
        1 2025-03-21                   0.997261                0.000  MULTI_ADJ_CASH
        2 2025-05-09                   1.000000                0.235   CASH_DIVIDEND
        3 2025-05-09                   0.997588                0.000  MULTI_ADJ_CASH
        """
        ref_data_type = ref_data_type.upper()
        node = otq.RefData(ref_data_type=ref_data_type)
        graph = otq.GraphQuery(node)
        df = otp.run(graph,
                     symbols=f'{self.name}::{symbol}',
                     symbol_date=symbol_date,
                     start=start,
                     end=end,
                     date=date,
                     timezone=timezone,
                     context=self.context)
        return df


def databases(context=utils.default, derived=False, readable_only: bool = True) -> dict[str, DB]:
    """
    Gets all available databases in the ``context``

    Parameters
    ----------
    context: str, optional
        Context to run the query.
        If not set then default :py:attr:`context<onetick.py.configuration.Config.context>` is used.
        See :ref:`guide about switching contexts <switching contexts>` for examples.
    derived: bool, dict
        If False (default) then derived databases are not returned.
        Otherwise derived databases names are added to the result after the non-derived databases.
        If set to dict then its items used as parameters to :py:func:`~onetick.py.derived_databases`.
        If set to True then default parameters for :py:func:`~onetick.py.derived_databases` are used.
    readable_only: bool
        If set to True (default), then return only the databases with read-access for the current user.
        Otherwise return all databases visible from the current process.

    See also
    --------
    | **SHOW_DB_LIST** OneTick event processor
    | **ACCESS_INFO** OneTick event processor
    | :py:func:`derived_databases`

    Returns
    -------
    dict
        Dict where keys are database names and values are :class:`DB <onetick.py.db._inspection.DB>` objects
        with ``context`` specified.
    """
    if readable_only:
        node = (
            otq.AccessInfo(info_type='DATABASES', show_for_all_users=False, deep_scan=True).tick_type('ANY')
            >> otq.Passthrough('DB_NAME,READ_ACCESS')
            >> otq.WhereClause(where='READ_ACCESS = 1')
        )
    else:
        node = (
            otq.ShowDbList().tick_type('ANY')
            >> otq.Passthrough('DATABASE_NAME')
        )
    dbs = otp.run(node,
                  symbols='LOCAL::',
                  # start and end times don't matter for this query, use some constants
                  start=db_constants.DEFAULT_START_DATE,
                  end=db_constants.DEFAULT_END_DATE,
                  context=context)

    # WebAPI returns empty DataFrame (no columns) if there are no databases
    if len(dbs) == 0:
        return {}

    db_list = list(dbs['DB_NAME'] if readable_only else dbs['DATABASE_NAME'])
    db_dict = {db_name: DB(db_name, context=context) for db_name in db_list}
    if derived:
        kwargs = derived if isinstance(derived, dict) else {}
        kwargs.setdefault('context', context)
        db_dict.update(
            derived_databases(**kwargs)
        )
    return db_dict


def derived_databases(
    context=utils.default,
    start=None, end=None,
    selection_criteria='all',
    db=None,
    db_discovery_scope='query_host_only',
) -> dict[str, DB]:
    """
    Gets available derived databases.

    Parameters
    ----------
    context: str, optional
        Context to run the query.
        If not set then default :py:attr:`context<onetick.py.configuration.Config.context>` is used.
        See :ref:`guide about switching contexts <switching contexts>` for examples.
    start: :py:class:`otp.datetime <onetick.py.datetime>`, optional
        If both ``start`` and ``end`` are set, then listing databases in this range only.
        Otherwise list databases from all configured time ranges for databases.

        If ``db`` is set, then
        :py:attr:`otp.config.default_start_time <onetick.py.configuration.Config.default_start_time>`
        is used by default.
    end: :py:class:`otp.datetime <onetick.py.datetime>`, optional
        If both ``start`` and ``end`` are set, then listing databases in this range only.
        Otherwise list databases from all configured time ranges for databases.

        If ``db`` is set, then
        :py:attr:`otp.config.default_end_time <onetick.py.configuration.Config.default_end_time>` is used by default.
    selection_criteria: str
        Possible values: *all*, *derived_from_current_db*, *direct_children_of_current_db*.
    db: str, optional
       Specifies database name if ``selection_criteria`` is set to
       *derived_from_current_db* or *direct_children_of_current_db*.
       Must be set in this case, otherwise does nothing.
    db_discovery_scope: str
        When *query_host_and_all_reachable_hosts* is specified,
        an attempt will be performed to get derived databases from all reachable hosts.
        When *query_host_only* is specified,
        only derived databases from the host on which the query is performed will be returned.

    See also
    --------
    **SHOW_DERIVED_DB_LIST** OneTick event processor

    Returns
    -------
    dict
        Dict where keys are database names and values are :class:`DB <onetick.py.db._inspection.DB>` objects
        with ``context`` specified.
    """
    if start and end:
        time_range = otq.ShowDerivedDbList.TimeRange.QUERY_TIME_INTERVAL
    else:
        if db is None:
            # start and end times don't matter in this case, use some constants
            start = db_constants.DEFAULT_START_DATE
            end = db_constants.DEFAULT_END_DATE
        else:
            start = otp.config.default_start_time
            end = otp.config.default_end_time
        time_range = otq.ShowDerivedDbList.TimeRange.CONFIGURED_TIME_INTERVAL

    selection_criteria = getattr(otq.ShowDerivedDbList.SelectionCriteria, selection_criteria.upper())
    db_discovery_scope = getattr(otq.ShowDerivedDbList.DbDiscoveryScope, db_discovery_scope.upper())

    if selection_criteria != otq.ShowDerivedDbList.SelectionCriteria.ALL and not db:
        raise ValueError(f"Parameter 'db' must be set when parameter 'selection_criteria' is {selection_criteria}")

    ep = otq.ShowDerivedDbList(
        time_range=time_range,
        selection_criteria=selection_criteria,
        db_discovery_scope=db_discovery_scope,
    )
    ep = ep.tick_type('ANY')
    db = db or 'LOCAL'
    dbs = otp.run(ep, symbols=f'{db}::', start=start, end=end, context=context)
    if len(dbs) == 0:
        return {}
    db_list = list(dbs['DERIVED_DB_NAME'])
    return {db_name: DB(db_name, context=context) for db_name in db_list}



import datetime as dt
import os
import io
import string

from functools import partial
from typing import Optional, Union, Dict

import onetick.py as otp
from onetick.py.otq import otq
import pandas as pd

from onetick.py.core._source._symbol_param import _SymbolParamSource
from onetick.py.core.source import Source

from .. import types as ott
from .. import utils, configuration
from ..core import _csv_inspector

from .common import default_date_converter, to_timestamp_nanos, update_node_tick_type
from .ticks import Ticks


def CSV(  # NOSONAR
    filepath_or_buffer=None,
    timestamp_name: Optional[str] = "Time",
    first_line_is_title: bool = True,
    names: Optional[list] = None,
    dtype: Optional[dict] = None,
    converters: Optional[dict] = None,
    order_ticks=False,
    drop_index=True,
    change_date_to=None,
    auto_increase_timestamps=True,
    db='LOCAL',
    field_delimiter=',',
    handle_escaped_chars=False,
    quote_char='"',
    timestamp_format: Optional[Union[str, Dict[str, str]]] = None,
    file_contents: Optional[str] = None,
    **kwargs,
):
    """
    Construct source based on CSV file.

    There are several steps determining column types.

    1. Initially, all column treated as ``str``.
    2. If column name in CSV title have format ``type COLUMNNAME``,
       it will change type from ``str`` to specified type.
    3. All column type are determined automatically from its data.
    4. You could override determined types in ``dtype`` argument explicitly.
    5. ``converters`` argument is applied after ``dtype`` and could also change column type.

    NOTE: Double quotes are not supported in CSV files for escaping quotes in strings,
    you should use escape character ``\\`` before the quote instead,
    for example: ``"I'm a string with a \\"quotes\\" inside"``. And then set `handle_escaped_chars=True`.

    Parameters
    ----------
    filepath_or_buffer: str, os.PathLike, FileBuffer, optional
        Path to CSV file or :class:`file buffer <FileBuffer>`. If None value is taken through symbol.
        When taken from symbol, symbol must have ``LOCAL::`` prefix.
        In that case you should set the columns otherwise schema will be empty.
    timestamp_name: str, default "Time"
        Name of TIMESTAMP column used for ticks. Used only if it is exists in CSV columns, otherwise ignored.
        Output data will be sorted by this column.
    first_line_is_title: bool
        Use first line of CSV file as a source for column names and types.
        If CSV file is started with # symbol, this parameter **must** be ``True``.

        - If ``True``, column names are inferred from the first line of the file,
          it is not allowed to have empty name for any column.

        - If ``False``, first line is processed as data, column names will be COLUMN_1, ..., COLUMN_N.
          You could specify column names in ``names`` argument.

    names: list, optional
        List of column names to use, or None.
        Length must be equal to columns number in file.
        Duplicates in this list are not allowed.
    dtype: dict, optional
        Data type for columns, as dict of pairs {column_name: type}.
        Will convert column type from ``str`` to specified type, before applying converters.
    converters: dict, optional
        Dict of functions for converting values in certain columns. Keys are column names.
        Function must be valid callable with ``onetick.py`` syntax, example::

            converters={
                "time_number": lambda c: c.apply(otp.nsectime),
                "stock": lambda c: c.str.lower(),
            }

        Converters applied *after* ``dtype`` conversion.
    order_ticks: bool, optional
        If ``True`` and ``timestamp_name`` column are used, then source will order tick by time.
        Note, that if ``False`` and ticks are not ordered in sequence, then OneTick will raise Exception in runtime.
    drop_index: bool, optional
        if ``True`` and 'Index' column is in the csv file then this column will be removed.
    change_date_to: datetime, date, optional
        change date from a timestamp column to a specific date. Default is None, means not changing timestamp column.
    auto_increase_timestamps: bool, optional
        Only used if provided CSV file does not have a TIMESTAMP column. If ``True``, timestamps of loaded ticks
        would start at ``start_time`` and on each next tick, would increase by 1 millisecond.
        If ``False``, timestamps of all loaded ticks would be equal to ``start_time``
    db: str, optional
        Name of a database to define a destination where the csv file will be transported for processing.
        ``LOCAL`` is default value that means OneTick will process it on the site where a query runs.
    field_delimiter: str, optional
        A character that is used to tokenize each line of the CSV file.
        For a tab character \t (back-slash followed by t) should be specified.
    handle_escaped_chars: bool, optional
        If set, the backslash char ``\\`` gets a special meaning and everywhere in the input text
        the combinations ``\\'``, ``\\"`` and ``\\\\`` are changed correspondingly by ``'``, ``"`` and ``\\``,
        which are processed then as regular chars.
        Besides, combinations like ``\\x??``, where ?-s are hexadecimal digits (0-9, a-f or A-F),
        are changed by the chars with the specified ASCII code.
        For example, ``\\x0A`` will be replaced by a newline character, ``\\x09`` will be replaced by tab, and so on.
        Default: False
    quote_char: str
        Character used to denote the start and end of a quoted item. Quoted items can include the delimiter,
        and it will be ignored. The same character cannot be marked both as the quote character and as the
        field delimiter. Besides, space characters cannot be used as quote.
        Default: " (double quotes)
    timestamp_format: str or dict
        Expected format for ``timestamp_name`` and all other datetime columns.
        If dictionary is passed, then different format can be specified for each column.
        This format is expected when converting strings from csv file to ``dtype``.
        Default format is ``%Y/%m/%d %H:%M:%S.%J`` for :py:class:`~onetick.py.nsectime` columns and
        ``%Y/%m/%d %H:%M:%S.%q`` for :py:class:`~onetick.py.msectime` columns.
    file_contents: str
        Specify the contents of the csv file as string.
        Can be used instead of ``filepath_or_buffer`` parameter.

    See also
    --------
    **CSV_FILE_LISTING** OneTick event processor

    Examples
    --------
    Simple CSV file reading

    >>> data = otp.CSV(os.path.join(csv_path, "data.csv"))
    >>> otp.run(data)
                         Time          time_number      px side
    0 2003-12-01 00:00:00.000  1656690986953602371   30.89  Buy
    1 2003-12-01 00:00:00.001  1656667706281508365  682.88  Buy

    Read CSV file and get timestamp for ticks from specific field.
    You need to specify query start/end interval including all ticks.

    >>> data = otp.CSV(os.path.join(csv_path, "data.csv"),
    ...                timestamp_name="time_number",
    ...                converters={"time_number": lambda c: c.apply(otp.nsectime)},
    ...                start=otp.dt(2010, 8, 1),
    ...                end=otp.dt(2022, 9, 2))
    >>> otp.run(data)
                               Time      px side
    0 2022-07-01 05:28:26.281508365  682.88  Buy
    1 2022-07-01 11:56:26.953602371   30.89  Buy

    Path to csv can be passed via symbol with `LOCAL::` prefix:

    >>> data = otp.CSV()
    >>> otp.run(data, symbols=f"LOCAL::{os.path.join(csv_path, 'data.csv')}")
                         Time          time_number      px side
    0 2003-12-01 00:00:00.000  1656690986953602371   30.89  Buy
    1 2003-12-01 00:00:00.001  1656667706281508365  682.88  Buy

    Field delimiters can be set via ``field_delimiters`` parameter:

    >>> data = otp.CSV(os.path.join(csv_path, 'data_diff_delimiters.csv'),
    ...                field_delimiter=' ',
    ...                first_line_is_title=False)
    >>> otp.run(data)
                         Time COLUMN_0 COLUMN_1
    0 2003-12-01 00:00:00.000      1,2        3
    1 2003-12-01 00:00:00.001        4      5,6

    Quote char can be set via ``quote_char`` parameter:

    >>> data = otp.CSV(os.path.join(csv_path, 'data_diff_quote_chars.csv'),
    ...                quote_char="'",
    ...                first_line_is_title=False)
    >>> otp.run(data)
                         Time COLUMN_0 COLUMN_1
    0 2003-12-01 00:00:00.000     1,"2       3"
    1 2003-12-01 00:00:00.001       "1     2",3

    Use parameter ``file_contents`` to read the data from string:

    >>> data = otp.CSV(file_contents=os.linesep.join([
    ...     'A,B,C',
    ...     '1,f,3.3',
    ...     '2,g,4.4',
    ... ]))
    >>> otp.run(data)
                         Time  A  B    C
    0 2003-12-01 00:00:00.000  1  f  3.3
    1 2003-12-01 00:00:00.001  2  g  4.4
    """
    csv_source = _CSV(
        filepath_or_buffer=filepath_or_buffer,
        timestamp_name=timestamp_name,
        first_line_is_title=first_line_is_title,
        names=names,
        dtype=dtype,
        converters=converters,
        order_ticks=order_ticks,
        drop_index=drop_index,
        change_date_to=change_date_to,
        auto_increase_timestamps=auto_increase_timestamps,
        db=db,
        field_delimiter=field_delimiter,
        handle_escaped_chars=handle_escaped_chars,
        quote_char=quote_char,
        timestamp_format=timestamp_format,
        file_contents=file_contents,
        **kwargs,
    )
    csv_source = csv_source.sort(csv_source['Time'])
    return otp.merge([csv_source, otp.Empty(db=db)])


class _CSV(Source):
    _PROPERTIES = Source._PROPERTIES + [
        "_dtype",
        "_names",
        "_columns",
        "_columns_with_bool_replaced",
        "_forced_title",
        "_default_types",
        "_has_time",
        "_to_drop",
        "_start",
        "_end",
        "_ep_fields",
        "_symbols",
        "_field_delimiter",
        "_converters",
        "_order_ticks",
        "_auto_increase_timestamps",
        "_db",
        "_drop_index",
        "_change_date_to",
        "_timestamp_name",
        "_filepath_or_buffer",
        "_first_line_is_title",
        "_handle_escaped_chars",
        "_quote_char",
        "_timestamp_format",
        "_file_contents",
    ]

    def __init__(self,
                 filepath_or_buffer=None,
                 timestamp_name: Optional[str] = "Time",
                 first_line_is_title: bool = True,
                 names: Optional[list] = None,
                 dtype: Optional[dict] = None,
                 converters: Optional[dict] = None,
                 order_ticks=False,
                 drop_index=True,
                 change_date_to=None,
                 auto_increase_timestamps=True,
                 db='LOCAL',
                 field_delimiter=',',
                 handle_escaped_chars=False,
                 quote_char='"',
                 timestamp_format: Optional[Union[str, Dict[str, str]]] = None,
                 file_contents: Optional[str] = None,
                 **kwargs):

        self._dtype = dtype or {}
        self._names = names
        self._converters = converters or {}
        if (len(field_delimiter) != 1 and field_delimiter != '\t') or field_delimiter == '"' or field_delimiter == "'":
            raise ValueError(f'`field_delimiter` can be single character (except quotes) '
                             f'or "\t" but "{field_delimiter}" was passed')
        self._field_delimiter = field_delimiter
        if len(quote_char) > 1:
            raise ValueError(f'quote_char should be single char but `{quote_char}` was passed')
        if self._field_delimiter == quote_char:
            raise ValueError(f'`{self._field_delimiter}` is both field_delimiter and quote_char')
        if quote_char in string.whitespace:
            raise ValueError('Whitespace can not be a quote_char')
        self._quote_char = quote_char
        self._order_ticks = order_ticks
        self._auto_increase_timestamps = auto_increase_timestamps
        self._db = db
        self._drop_index = drop_index
        self._change_date_to = change_date_to
        self._timestamp_name = timestamp_name
        self._filepath_or_buffer = filepath_or_buffer
        self._first_line_is_title = first_line_is_title
        self._handle_escaped_chars = handle_escaped_chars
        self._timestamp_format = timestamp_format
        self._file_contents = file_contents

        if self._try_default_constructor(**kwargs):
            return

        if self._filepath_or_buffer is not None and self._file_contents is not None:
            raise ValueError("Parameters 'filepath_or_buffer' and 'file_contents' can't be set at the same time.")

        need_to_parse_file = (
            self._file_contents is not None or
            self._filepath_or_buffer is not None and not isinstance(self._filepath_or_buffer, _SymbolParamSource)
        )
        if need_to_parse_file:
            self._columns, self._default_types, self._forced_title, self._symbols = self._parse_file()
        else:
            self._filepath_or_buffer = None
            names = self._names or []
            self._columns = {name: str for name in names}
            self._default_types = {}
            # we don't know it is actually forced, but otherwise we would ignore the first not commented-out line
            self._forced_title = self._first_line_is_title
            self._symbols = None

        self._check_time_column()

        for t in self._dtype:
            if t not in self._columns:
                raise ValueError(f"dtype '{t}' not found in columns list")
            self._columns[t] = self._dtype[t]

        self._ep_fields = ",".join(
            f'{ott.type2str(dtype)} {column}' if issubclass(dtype, otp.string) else column
            for column, dtype in self._columns.items()
        )

        self._to_drop = self._get_to_drop()
        self._has_time, self._start, self._end = self._get_start_end(**kwargs)

        self._columns_with_bool_replaced = dict((n, c if c != bool else float) for n, c in self._columns.items())

        super().__init__(
            _symbols=self._symbols,
            _start=self._start,
            _end=self._end,
            _base_ep_func=self.base_ep,
            schema=self._columns_with_bool_replaced,
        )

        # fake run converters to set proper schema
        if self._converters:
            for column, converter in self._converters.items():
                self.schema[column] = converter(self[column]).dtype

        if self._has_time and self._timestamp_name in self.schema:
            if self.schema[self._timestamp_name] not in [ott.nsectime, ott.msectime]:
                raise ValueError(f"CSV converter for {self._timestamp_name} is converting to "
                                 f"{self.schema[timestamp_name]} type, but expected resulted type is "
                                 f"ott.msectime or ott.nsectime")

        # remove timestamp_name column, if we use it as TIMESTAMP source
        if self._has_time and self._timestamp_name != "Time":
            del self[self._timestamp_name]

    def _check_time_column(self):
        if "TIMESTAMP" in self._columns:
            raise ValueError(
                "It is not allowed to have 'TIMESTAMP' columns, because it is reserved name in OneTick"
            )

        if "Time" in self._columns and self._timestamp_name != "Time":
            raise ValueError(
                "It is not allowed to have 'Time' column not used as timestamp field."
            )

    def _get_to_drop(self):
        to_drop = []
        if "TICK_STATUS" in self._columns:
            del self._columns["TICK_STATUS"]
            to_drop.append("TICK_STATUS")

        if "Index" in self._columns and self._drop_index:
            del self._columns["Index"]
            to_drop.append("Index")
        return to_drop

    def _get_start_end(self, **kwargs):
        start = kwargs.get("start", utils.adaptive)
        end = kwargs.get("end", utils.adaptive)

        has_time = False
        if self._timestamp_name in self._columns:
            has_time = True

            # remove to resolve exception in Source.__init__
            if self._timestamp_name == "Time":
                del self._columns["Time"]

        # redefine start/end time for change_date_to
        if self._change_date_to:
            start = dt.datetime(self._change_date_to.year, self._change_date_to.month, self._change_date_to.day)
            end = ott.next_day(start)
        return has_time, start, end

    def _parse_file(self):
        """
        This function finds the file and get columns names, default types and checks if first line is title via pandas.
        Is also sets the correct value for symbols.
        """
        obj_to_inspect = self._filepath_or_buffer
        if isinstance(obj_to_inspect, utils.FileBuffer):
            obj_to_inspect = io.StringIO(obj_to_inspect.get())
        if self._file_contents is not None:
            obj_to_inspect = io.StringIO(self._file_contents)

        if isinstance(obj_to_inspect, str) and not os.path.exists(obj_to_inspect):
            # if not found, probably, CSV file is located in OneTick CSV_FILE_PATH, check it for inspect_by_pandas()
            csv_paths = otp.utils.get_config_param(os.environ["ONE_TICK_CONFIG"], "CSV_FILE_PATH", default="")
            if csv_paths:
                for csv_path in csv_paths.split(","):
                    csv_path = os.path.join(csv_path, obj_to_inspect)
                    if os.path.exists(csv_path):
                        obj_to_inspect = csv_path
                        break

        columns, default_types, forced_title = _csv_inspector.inspect_by_pandas(
            obj_to_inspect,
            self._first_line_is_title,
            self._names,
            self._field_delimiter,
            self._quote_char,
        )
        if isinstance(self._filepath_or_buffer, utils.FileBuffer) or self._file_contents is not None:
            symbols = 'DUMMY'
        else:
            # str, because there might passed an os.PathLike object
            symbols = str(obj_to_inspect)
        return columns, default_types, forced_title, symbols

    def _get_timestamp_format(self, column_name, dtype):
        if dtype not in (ott.nsectime, ott.msectime):
            raise ValueError(f"Wrong value for parameter 'dtype': {dtype}")
        if self._timestamp_format is None:
            # by default we parse timestamp_name into TIMESTAMP field
            # from typical/default Time format from OneTick dump
            if dtype is ott.nsectime:
                return '%Y/%m/%d %H:%M:%S.%J'
            else:
                return '%Y/%m/%d %H:%M:%S.%q'
        if isinstance(self._timestamp_format, dict):
            return self._timestamp_format[column_name]
        return self._timestamp_format

    def base_ep(self):
        # initialize Source and set schema to columns.
        file_contents = ''
        columns_to_drop = self._to_drop.copy()

        if isinstance(self._filepath_or_buffer, utils.FileBuffer):
            file_contents = self._filepath_or_buffer.get()
        if self._file_contents is not None:
            file_contents = self._file_contents

        csv = Source(
            otq.CsvFileListing(
                field_delimiters=f"'{self._field_delimiter}'",
                time_assignment="_START_TIME",
                # we use EP's first_line_is_title only when file path is passed through symbol
                # otherwise we don't use EP's first_line_is_title, because EP raise error on empty column name,
                # and we explicitly define name for such columns in FIELDS arg.
                # but if first line started with # (forced_title=True), then this param ignored :(
                first_line_is_title=(self._filepath_or_buffer is None and
                                     self._file_contents is None and
                                     self._first_line_is_title),
                fields=self._ep_fields,
                file_contents=file_contents,
                handle_escaped_chars=self._handle_escaped_chars,
                quote_chars=f"'{self._quote_char}'",
            ),
            schema=self._columns_with_bool_replaced,
        )

        if self._first_line_is_title and not self._forced_title:
            # remove first line with titles for columns.
            csv.sink(otq.DeclareStateVariables(variables="long __TICK_INDEX=0"))
            csv.sink(otq.PerTickScript("STATE::__TICK_INDEX = STATE::__TICK_INDEX + 1;"))
            csv.sink(otq.WhereClause(discard_on_match=False, where="STATE::__TICK_INDEX > 1"))

        # set tick type to ANY
        update_node_tick_type(csv, "ANY", self._db)

        # check whether need to update types, because if column type is not specified in header
        # then by default column has string type in OneTick
        update_columns = {}
        for name, dtype in self._columns.items():
            if not issubclass(dtype, str) and name not in self._default_types:
                update_columns[name] = dtype

        for name, dtype in update_columns.items():
            if dtype is int:
                # BE-142 - workaround for converting string to int
                # OneTick first convert string to float, and then to int, which leeds to losing precision
                csv.sink(otq.AddField(field=f"_TMP_{name}", value="atol(" + name + ")"))
                csv.sink(otq.Passthrough(fields=name, drop_fields=True))
                csv.sink(otq.AddField(field=f"{name}", value=f"_TMP_{name}"))
                csv.sink(otq.Passthrough(fields=f"_TMP_{name}", drop_fields=True))
            elif dtype is float:
                csv.sink(otq.UpdateField(field=name, value="atof(" + name + ")"))
            elif dtype is ott.msectime:
                timestamp_format = self._get_timestamp_format(name, dtype)
                csv.sink(otq.UpdateField(field=name,
                                         value=f'time_format("{timestamp_format}",0,_TIMEZONE)',
                                         where=name + '=""'))
                csv.sink(otq.UpdateField(field=name, value=f'parse_time("{timestamp_format}",{name},_TIMEZONE)'))
            elif dtype is ott.nsectime:
                timestamp_format = self._get_timestamp_format(name, dtype)
                csv.sink(otq.UpdateField(field=name,
                                         value=f'time_format("{timestamp_format}",0,_TIMEZONE)',
                                         where=name + '=""'))
                # TODO: this is the logic from otp.Source._update_field,
                # we should use _Source methods here or refactor
                csv.sink(otq.AddField(field=f"_TMP_{name}",
                                      value=f'parse_nsectime("{timestamp_format}",{name},_TIMEZONE)'))
                csv.sink(otq.Passthrough(fields=name, drop_fields=True))
                csv.sink(otq.AddField(field=name, value=f"_TMP_{name}"))
                csv.sink(otq.Passthrough(fields=f"_TMP_{name}", drop_fields=True))
            elif dtype is bool:
                csv.sink(otq.UpdateField(field=name, value="CASE(" + name + ", 'true', 1.0, 0.0)"))
            else:
                raise TypeError(f"Unsupported type '{dtype}'")

        # run converters
        if self._converters:
            for column, converter in self._converters.items():
                if csv[column].dtype is not otp.nsectime and converter(csv[column]).dtype is otp.nsectime:
                    # workaround for resolve bug on column type changing:
                    # https://onemarketdata.atlassian.net/browse/PY-416
                    csv[f'_T_{name}'] = converter(csv[column])
                    del csv[column]
                    csv[column] = csv[f'_T_{name}']
                    del csv[f'_T_{name}']
                else:
                    csv[column] = converter(csv[column])

        if self._has_time:
            # if timestamp_name column is defined in the csv, then apply tick time adjustment

            if self._timestamp_name in self._converters:
                # we assume that if timestamp_name field in converters,
                # then it is already converted to otp.dt
                csv.sink(
                    otq.UpdateField(
                        field="TIMESTAMP",
                        value=self._timestamp_name,
                        allow_unordered_output_times=True,
                    )
                )
            else:
                if self._change_date_to:
                    self._change_date_to = self._change_date_to.strftime("%Y/%m/%d")
                    csv.sink(otq.UpdateField(field="Time",
                                             value=f'"{self._change_date_to}" + substr({self._timestamp_name}, 10)'))

                timestamp_format = self._get_timestamp_format(self._timestamp_name, otp.nsectime)
                csv.sink(
                    otq.UpdateField(
                        field="TIMESTAMP",
                        value=f'parse_nsectime("{timestamp_format}", {self._timestamp_name}, _TIMEZONE)',
                        allow_unordered_output_times=True,
                    )
                )

            # drop source timestamp_name field in favor of new TIMESTAMP field
            columns_to_drop.append(self._timestamp_name)
        elif self._auto_increase_timestamps:
            # default time for ticks are increasing from 0
            csv.sink(otq.DeclareStateVariables(variables="long __TIMESTAMP_INC__ = 0"))
            csv.sink(otq.UpdateField(
                field="TIMESTAMP",
                value='DATEADD("millisecond",STATE::__TIMESTAMP_INC__,TIMESTAMP,_TIMEZONE)'))
            csv.sink(otq.UpdateField(field="STATE::__TIMESTAMP_INC__", value="STATE::__TIMESTAMP_INC__ + 1"))

        if self._order_ticks:
            csv.sort('TIMESTAMP', inplace=True)

        if columns_to_drop:
            csv.sink(otq.Passthrough(fields=",".join(columns_to_drop), drop_fields="True"))

        return csv


def LocalCSVTicks(path,  # NOSONAR
                  start=utils.adaptive,
                  end=utils.adaptive,
                  date_converter=default_date_converter,
                  additional_date_columns=None,
                  converters=None,
                  tz=None,
                  ):
    """
    Loads ticks from csv file, and creating otp.Ticks object from them

    Parameters
    ----------
    path: str
        Absolute path to csv file
    start: datetime object
        Start of the query interval
    end: datetime object
        End of the query interval
    date_converter:
        A converter from string to datetime format, by default used only to TIMESTAMP column
    additional_date_columns:
        Other columns to convert to datetime format
    converters:
        Non default converters to columns from strings
    tz:
        timezone

    Returns
    -------
    otp.Ticks
    """
    if tz is None:
        tz = configuration.config.tz

    c = {'TIMESTAMP': partial(to_timestamp_nanos, date_converter=date_converter, tz=tz)}
    if converters is not None:
        c.update(converters)
    if additional_date_columns is not None:
        c.update({column: partial(to_timestamp_nanos,
                                  date_converter=date_converter,
                                  tz=tz,
                                  ) for column in additional_date_columns})
    df = pd.read_csv(path, converters=c)
    df['TS_'] = df['TIMESTAMP']
    df['SYMBOL_NAME'] = df['#SYMBOL_NAME']
    d = df.to_dict(orient='list')
    del d['TIMESTAMP']
    del d['#SYMBOL_NAME']

    ticks = Ticks(d, start=start, end=end)
    ticks['TIMESTAMP'] = ticks['TS_']
    ticks = ticks.drop('TS_')

    return ticks



from onetick.py.core._internal._op_utils.every_operand import every_operand
from onetick.py.core._source._symbol_param import _SymbolParamColumn
from onetick.py.core.column import _Column


def is_const(operation):
    for op in every_operand(operation):
        if isinstance(op, _Column) and not isinstance(op, _SymbolParamColumn):
            return False
    return True



import os
import re
import html
import graphviz as gv
from collections import defaultdict, deque
from datetime import datetime
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Deque, Dict, List, Optional, Set, Tuple, Union

from onetick.py.utils import TmpFile


EPS_WITH_QUERIES = {
    "eval": (0, "expression"),
    "eval_expression": (0, "expression"),
    "join_with_query": (0, "otq_query"),
    "nested_otq": (0, "otq_name"),
    "join_with_collection_summary": (1, "otq_query"),
    "modify_state_var_from_query": (1, "otq_query"),
    "create_cache": (1, "otq_file_path"),
    "read_cache": (4, "create_cache_query"),
    "generic_aggregation": (0, "query_name"),
}

IF_ELSE_EPS = {
    "CHARACTER_PRESENT", "CORRECT_TICK_FILTER", "PRIMARY_EXCH", "REGEX_MATCHES", "SKIP_BAD_TICK", "TIME_FILTER",
    "TRD_VS_MID", "TRD_VS_QUOTE", "UPTICK", "VALUE_COMPARE", "VALUE_PRESENT", "VOLUME_LIMIT", "WHERE_CLAUSE",
}


@dataclass
class NestedQuery:
    name: str
    raw_string: str
    query: Optional[str] = field(default=None)
    expression: Optional[str] = field(default=None)
    file_path: Optional[str] = field(default=None)
    args: list = field(default_factory=list)
    kwargs: dict = field(default_factory=dict)
    is_local: bool = field(default=False)

    def to_string(self):
        if self.is_local:
            if self.file_path:
                raise ValueError("Nested query from file couldn't be local")

            if self.expression:
                return self.expression
            else:
                return self.query
        else:
            return "::".join(i for i in [self.file_path, self.query] if i)


@dataclass
class EP:
    name: str
    raw_string: str
    args: list = field(default_factory=list)
    kwargs: dict = field(default_factory=dict)


@dataclass
class IfElseEP(EP):
    if_nodes: Set[str] = field(default_factory=set)
    else_nodes: Set[str] = field(default_factory=set)


@dataclass
class Node:
    ep: Union[EP, NestedQuery, None]
    id: str
    query: str
    tick_type: Optional[str] = field(default=None)
    labels: Dict[str, str] = field(default_factory=dict)
    config: dict = field(default_factory=dict)
    params: dict = field(default_factory=dict)
    sinks: List[str] = field(default_factory=list)
    symbols: list = field(default_factory=list)


@dataclass
class Query:
    name: str
    graph: str
    nodes: Dict[str, Node] = field(default_factory=dict)
    roots: list = field(default_factory=list)
    leaves: list = field(default_factory=list)
    symbols: list = field(default_factory=list)
    config: dict = field(default_factory=dict)
    params: dict = field(default_factory=dict)
    depends: Set[Tuple[Optional[str], Optional[str]]] = field(default_factory=set)

    def get_id(self, prefix: Optional[str] = "cluster"):
        if prefix:
            prefix = f"{prefix}__"
        else:
            prefix = ""

        graph = self.graph.replace(":", "_")

        return f"{prefix}{graph}__{self.name}"


@dataclass
class Graph:
    file_path: str
    config: dict = field(default_factory=dict)
    queries: Dict[str, Query] = field(default_factory=dict)

    def has_query(self, query):
        return query in self.queries


class GraphStorage(dict):
    def get_query(self, otq_file: Optional[str], query_name: Optional[str]) -> Optional[Query]:
        if not query_name or not otq_file:
            return None

        if otq_file not in self or query_name not in self[otq_file].queries:
            return None

        return self[otq_file].queries[query_name]

    def get_query_unique_id(self, query: Union[str, Query, NestedQuery], graph: Optional[str] = None) -> Optional[str]:
        query_obj = None
        if isinstance(query, Query):
            query_obj = query
        elif isinstance(query, NestedQuery):
            if not query.is_local or query.expression:
                raise RuntimeError("Couldn't get id for non-local or expression-based NestedQuery")

            query_obj = self.get_query(query.file_path, query.query)
        elif isinstance(query, str):
            if not graph:
                raise ValueError("`graph` with path to otq file is required for `str` query")

            query_obj = self.get_query(graph, query)
        else:
            raise RuntimeError(f"Unsupported query type: {type(query)}")

        if not query_obj:
            return None
        else:
            return query_obj.get_id()


class GVTable:
    def __init__(self, border=0, cellborder=1, cellspacing=0, attrs: Optional[dict] = None, auto_colspan=True):
        """
        Generate HTML tables for Graphviz

        Attributes for each row or cell can be set by passing `attrs` attribute to `GVTable.row` or `GVTable.cell`.
        Accoringly `attrs` params have reverse prioriry: higher for cells, less for rows.

        Parameters
        ----------
        border: int
            Value of `BORDER` attribute of table HTML element
        cellborder: int
            Value of `CELLBORDER` attribute of table HTML element
        cellspacing: int
            Value of `CELLSPACING` attribute of table HTML element
        attrs: dict
            HTML attributes to apply to table element.
        auto_colspan: bool
            If set True, then last cell in each row automaticly fills table width,
            if `colspan` attribute not set for this cell.

        Examples
        --------

        Simple two rows table:

        >>> table = otp.utils.render.GVTable()
        >>> table.row(["cell_1_1", "cell_1_2"])  # doctest: +SKIP
        >>> table.row(["cell_2_1", "cell_2_2"])  # doctest: +SKIP
        >>> table_html = str(table)  # doctest: +SKIP
        """
        self.rows: List[Tuple[List[Tuple[Union[List[str], str], dict]], dict]] = []
        self.attrs = {
            "border": border,
            "cellborder": cellborder,
            "cellspacing": cellspacing,
        }
        if attrs:
            self.attrs.update(attrs)

        self.auto_colspan = auto_colspan
        self.max_cols = 0

    def cell(self, data: list, attrs: Optional[dict] = None):
        """
        Append cell in the last row
        """
        if len(self) == 0:
            raise RuntimeError("No rows in table")

        row, row_attrs = self.rows[-1]

        for cell_data in data:
            cell_attrs = row_attrs.copy()

            if attrs:
                cell_attrs.update(attrs)

            if isinstance(cell_data, tuple):
                cell, _cell_attrs = cell_data
                if _cell_attrs:
                    cell_attrs.update(_cell_attrs)
            else:
                cell = cell_data

            row.append((cell, cell_attrs))

        self.max_cols = max(self.max_cols, len(row))

        return self

    def row(self, data: list, attrs: Optional[dict] = None):
        self.rows.append(([], attrs if attrs else {}))
        self.cell(data)

        return self

    def __len__(self):
        return len(self.rows)

    def __str__(self):
        tags = []

        attrs = " ".join([f"{k.upper()}=\"{v}\"" for k, v in self.attrs.items()])
        tags.append(f"<TABLE {attrs}>")

        for row, row_attrs in self.rows:
            col_count = len(row)
            tags.append("<TR>")

            for i in range(col_count):
                cell, cell_attrs = row[i]

                if (
                    self.auto_colspan and col_count - 1 == i and
                    len(row) < self.max_cols and "colspan" not in cell_attrs
                ):
                    cell_attrs["colspan"] = self.max_cols - col_count + 1

                attrs = " ".join([f"{k.upper()}=\"{v}\"" for k, v in cell_attrs.items()])
                if not isinstance(cell, list):
                    cell = [cell]

                cell_str = "<BR/>".join(cell)
                tags.append(f"<TD {attrs}>{cell_str}</TD>")

            tags.append("</TR>")

        tags.append("</TABLE>")

        return "<" + "".join(tags) + ">"


def _posix_path(path: str):
    return path.replace(os.sep, "/")


def _get_node_unique_id(node: Union[Node, str], query: Query):
    if isinstance(node, str):
        node = query.nodes[node]

    return f"{query.get_id(prefix=None)}__{node.id}"


def _save_param(storage, key, value):
    if key in storage:
        if not isinstance(storage[key], list):
            storage[key] = [storage[key]]

        storage[key].append(value)
    else:
        storage[key] = value


def _is_local_query(f_path: Optional[str]) -> bool:
    return f_path in ["THIS", "___ME___"]


def _parse_function_params(func_params: str) -> Tuple[list, dict]:
    def save_param(_key, _value, _args, _kwargs):
        if _key:
            _kwargs[_key.lower()] = (_key, _value)
        else:
            if _kwargs:
                raise RuntimeError("Positional argument could not be after keyword argument")

            _args.append(_value)

    args: list = []
    kwargs: dict = {}
    key = ""
    str_buffer: list = []
    in_quotes = None
    escape_next = False

    if not func_params:
        return args, kwargs

    for ch in func_params:
        if escape_next:
            escape_next = False
            str_buffer.append(ch)
        elif ch == "\\":
            escape_next = True
        elif in_quotes:
            if ch == in_quotes:
                in_quotes = None
            else:
                str_buffer.append(ch)
        else:
            if ch in "\"\'":
                in_quotes = ch
            elif ch.isspace():
                continue
            elif ch == "=":
                key = "".join(str_buffer)
                str_buffer.clear()
            elif ch == ",":
                save_param(key, "".join(str_buffer), args, kwargs)
                str_buffer.clear()
                key = ""
            else:
                str_buffer.append(ch)

    if in_quotes:
        raise ValueError("`func_params` unclosed quote")

    if str_buffer:
        save_param(key, "".join(str_buffer), args, kwargs)
        str_buffer.clear()

    return args, kwargs


def _parse_function(expression: str) -> Tuple[Optional[str], list, dict]:
    # EP_NAME(PARAM_NAME=PARAM_VALUE,...)
    # [a-zA-Z_:] is EP_NAME, can contain letters, underscore and colon
    # [\s\S] is any symbol including newline (because . doesn't include newline by default)
    m = re.search(r"^([a-zA-Z_:]*)\s*\(([\s\S]*)\)\s*$", expression)

    if not m:
        return None, [], {}

    ep, params_str = m.groups()
    args, kwargs = _parse_function_params(params_str)

    return ep, args, kwargs


def _get_ep_from_str(ep_string: str) -> Tuple[str, list, dict]:
    ep, args, kwargs = _parse_function(ep_string)

    if not ep:
        ep = ep_string

    return ep, args, kwargs


def _parse_query_path(query_path: str) -> Union[Tuple[str, Optional[str]], List[str]]:
    query_path_splitted = query_path.rsplit("::", maxsplit=1)

    if len(query_path_splitted) == 1:
        return _posix_path(query_path_splitted[0]), None
    else:
        file_path, query = query_path_splitted

        return _posix_path(file_path), query


def _parse_ep(ep_string: str, parse_eval_from_params: bool = False) -> Union[EP, NestedQuery]:
    if ep_string.startswith("NESTED_OTQ"):
        query_path = " ".join(ep_string.split(" ")[1:])

        file_path, query = _parse_query_path(query_path)
        is_local = _is_local_query(file_path)

        return NestedQuery(
            name="NESTED_OTQ", raw_string=ep_string, query=query, file_path=None if is_local else file_path,
            is_local=is_local,
        )

    ep, args, kwargs = _get_ep_from_str(ep_string)

    if parse_eval_from_params:
        for param_name, param_value in kwargs.items():
            m = re.search(r"^(eval\([^)]+\)).*$", param_value[1], re.IGNORECASE)
            if not m:
                continue

            param_ep_str = m.group(1)

            try:
                param_ep = _parse_ep(param_ep_str, parse_eval_from_params=False)
                if isinstance(param_ep, NestedQuery):
                    kwargs[param_name] = (param_value[0], param_ep)
            except Exception:
                pass

    if ep.lower() in EPS_WITH_QUERIES:
        ep_description = EPS_WITH_QUERIES[ep.lower()]
        args_idx, kwargs_key = ep_description

        is_query_found = True

        if kwargs_key in kwargs:
            query_path = kwargs.pop(kwargs_key)[1]
        elif 0 <= args_idx < len(args):
            query_path = args.pop(args_idx)
        else:
            # don't do anything, just process as EP
            is_query_found = False

        if is_query_found:
            if query_path[0] in ["\"", "\'"] and query_path[0] == query_path[-1]:
                query_path = query_path[1:-1]

            file_path, query = _parse_query_path(query_path)

            if file_path and query:
                is_local = _is_local_query(file_path)
                return NestedQuery(
                    name=ep, raw_string=ep_string, query=query, file_path=None if is_local else file_path,
                    args=args, kwargs=kwargs, is_local=is_local,
                )
            else:
                return NestedQuery(
                    name=ep, raw_string=ep_string, expression=file_path, args=args, kwargs=kwargs, is_local=True,
                )

    if ep in IF_ELSE_EPS:
        return IfElseEP(name=ep, raw_string=ep_string, args=args, kwargs=kwargs)

    return EP(name=ep, raw_string=ep_string, args=args, kwargs=kwargs)


def _parse_security(value: str) -> Tuple[Union[str, EP, NestedQuery], str, bool]:
    is_security_active = True
    split_value = value.split()

    try:
        int(split_value[-1])
    except ValueError:
        # assume that third value is "No"
        is_security_active = False
        split_value.pop()

    security = " ".join(split_value[:-1])

    try:
        security_ep = _parse_ep(security)
    except ValueError:
        security_ep = None

    return (security_ep if security_ep else security), split_value[-1], is_security_active


def _move_between_dicts(source, output, key, func):
    if not isinstance(source[key], list):
        source[key] = [source[key]]

    output.update([func(k.split()) for k in source[key]])
    del source[key]


def _move_parameters(_from, _to):
    if "PARAMETER" in _from:
        _move_between_dicts(
            _from,
            _to,
            "PARAMETER",
            lambda x: (x[0], " ".join(x[1:]))
        )

    if "PARAMETER_MANDATORY" in _from:
        _move_between_dicts(
            _from,
            _to,
            "PARAMETER_MANDATORY",
            lambda x: (x[0], None)
        )


def _build_query_tree(query: Query):
    roots = {*query.nodes.keys()}
    leaves = {*query.nodes.keys()}

    for node_id, node in query.nodes.items():
        _move_parameters(node.config, node.params)

        # save labels
        if "NESTED_INPUT" in node.config:
            node.labels["IN"] = node.config["NESTED_INPUT"]

        if "NESTED_OUTPUT" in node.config:
            node.labels["OUT"] = node.config["NESTED_OUTPUT"]

        if "SOURCE_DESCRIPTION" in node.config:
            descriptions = node.config["SOURCE_DESCRIPTION"]
            if isinstance(descriptions, str):
                descriptions = [descriptions]
            for description in descriptions:
                description = description.strip().split(" ")
                if len(description) > 1:
                    desc_node = description[0].split(".")[0]
                    labels = description[1].split(".")

                    if labels and desc_node in query.nodes:
                        if labels[0]:
                            query.nodes[desc_node].labels["IN"] = labels[0]

                        if labels[1]:
                            query.nodes[desc_node].labels["OUT"] = labels[1]

        if "SINK_DESCRIPTION" in node.config:
            description_path = node.config["SINK_DESCRIPTION"].strip().split(".")
            if len(description_path) > 1:
                desc_node = description_path[0]
                label = description_path[-1]

                if label and desc_node in query.nodes:
                    query.nodes[desc_node].labels["OUT"] = label

        nodes = []

        if "SINK" in node.config:
            sink_nodes = [(sink, True) for sink in node.config["SINK"].strip().split()]
            if sink_nodes:
                leaves.discard(node_id)

            nodes += sink_nodes
            del node.config["SINK"]

        if "SOURCE" in node.config:
            source_nodes = [(sink, False) for sink in node.config["SOURCE"].strip().split()]
            if source_nodes:
                roots.discard(node_id)

            nodes += source_nodes
            del node.config["SOURCE"]

        for source_node, is_sink_node in nodes:
            # just ignore other nodes in path
            source_node_path = source_node.split(".")
            source_node_id = source_node_path[0]

            if not is_sink_node and source_node_id in leaves:
                leaves.discard(source_node_id)

            if is_sink_node and source_node_id in roots:
                roots.discard(source_node_id)

            if source_node_id not in query.nodes:
                raise RuntimeError(f"Malformed otq file passed: node {source_node_id} not found in {query.name}")

            if is_sink_node:
                query.nodes[node_id].sinks.append(source_node_id)

                if isinstance(node.ep, IfElseEP):
                    if "IF" in source_node_path[1:]:
                        node.ep.if_nodes.add(source_node_id)

                    if "ELSE" in source_node_path[1:]:
                        node.ep.else_nodes.add(source_node_id)
            else:
                source_node = query.nodes[source_node_id]
                source_node.sinks.append(node_id)

                if isinstance(source_node.ep, IfElseEP):
                    if "IF" in source_node_path[1:]:
                        source_node.ep.if_nodes.add(node_id)

                    if "ELSE" in source_node_path[1:]:
                        source_node.ep.else_nodes.add(node_id)

    query.roots = list(roots)
    query.leaves = list(leaves)


def _save_dependency(obj, query: Query):
    if isinstance(obj, (EP, NestedQuery)):
        for kwarg_param in obj.kwargs.values():
            if isinstance(kwarg_param[1], NestedQuery):
                _save_dependency(kwarg_param[1], query)

        if isinstance(obj, NestedQuery) and not obj.expression:
            query.depends.add((obj.file_path, obj.query))


def _finalize_query(query: Query, graph: Graph):
    if not query:
        return

    if query.name == "_meta":
        graph.config = {k.upper(): v for k, v in query.config.items()}
        return

    if "SECURITY" in query.config:
        if not isinstance(query.config["SECURITY"], list):
            query.config["SECURITY"] = [query.config["SECURITY"]]

        for security in query.config["SECURITY"]:
            parsed_security = _parse_security(security)
            query.symbols.append(parsed_security)
            _save_dependency(parsed_security[0], query)

        del query.config["SECURITY"]

    _move_parameters(query.config, query.params)
    _build_query_tree(query)

    graph.queries[query.name] = query


def read_otq(path: str, parse_eval_from_params: bool = False) -> Optional[Graph]:
    if path.startswith("remote://") or not os.path.exists(path):
        return None

    graph = Graph(path)
    current_query = None

    with open(path, "r") as input_otq:
        tmp_line = ""
        for line in input_otq:
            line = line.rstrip()

            if line.endswith("\\"):
                tmp_line += f"{line[:-1]}\n"
                continue
            else:
                tmp_line += line

            line = tmp_line.strip()
            tmp_line = ""

            if not line:
                continue

            # found new query
            m = re.search(r"^\[(.*)\]$", line)
            if m:
                if current_query and current_query.config['TYPE'] == 'GRAPH':
                    _finalize_query(current_query, graph)

                query_name = m.groups()[0]
                current_query = Query(name=query_name, graph=path)
                continue

            if not current_query:
                continue

            if line.startswith("NODE"):
                prefix = r"NODE\D*?(\d+)"
            elif line.startswith("ROOT"):
                prefix = r"ROOT\D*?(\d*)"
            else:
                # other query param
                line_expr = line.split("=")
                param, value = line_expr[0], "=".join(line_expr[1:])
                param = param.strip()
                value = value.strip()

                _save_param(current_query.config, param, value)
                continue

            m = re.search(rf"^({prefix})(_([a-zA-Z_]*[0-9]*))?\s*=\s*([\s\S]*)$", line)
            if m:
                node_id, _, _, node_param, value = m.groups()

                if node_id not in current_query.nodes:
                    current_query.nodes[node_id] = Node(ep=None, id=node_id, query=current_query.name)

                if not node_param:
                    ep = _parse_ep(value, parse_eval_from_params=parse_eval_from_params)
                    _save_dependency(ep, current_query)
                    current_query.nodes[node_id].ep = ep
                elif node_param == "BIND_SECURITY":
                    security = _parse_security(value)
                    current_query.nodes[node_id].symbols.append(security)
                    _save_dependency(security[0], current_query)
                elif node_param == "TICK_TYPE":
                    current_query.nodes[node_id].tick_type = value
                else:
                    _save_param(current_query.nodes[node_id].config, node_param, value)

    if current_query:
        _finalize_query(current_query, graph)

    return graph


def truncate_param_value(ep: Any, param, value, line_limit: Optional[Tuple[int, int]] = None):
    if line_limit is None:
        return value

    height, width = line_limit
    if height < 0 or width < 0:
        raise ValueError("line_limit values should not be negative")

    if isinstance(ep, EP) and (
        ep.name == "PER_TICK_SCRIPT" and param.lower() == "script" or
        ep.name == "CSV_FILE_LISTING" and param.lower() == "file_contents"
    ):
        lines = [
            line if len(line) <= width or not width else line[:width] + "..."
            for line in value.split("\n")
        ]

        if height and len(lines) > height:
            lines = lines[:height] + ["..."]

        return "\n".join(lines)

    return value


def build_symbols(symbols, gr_nested, gr_static, graphs: GraphStorage, graph_node, reverse=False):
    table = GVTable()

    for symbol_data in symbols:
        symbol, _, _ = symbol_data

        if isinstance(symbol, NestedQuery):
            if symbol.query:
                if symbol.is_local:
                    # reversed directions here brakes everything
                    gr_nested.edge(
                        f"cluster__{symbol.query}__footer",
                        f"{graph_node}:symbols",
                        ltail=f"cluster__{symbol.query}",
                        style="dashed", constraint="false",
                    )
                    continue

                nested_cluster_id = graphs.get_query_unique_id(symbol.query, symbol.file_path)

                if nested_cluster_id:
                    gr_nested.edge(
                        f"{nested_cluster_id}__footer",
                        f"{graph_node}:symbols",
                        ltail=nested_cluster_id,
                        style="dashed", constraint="false",
                    )
                    continue

            query = symbol.to_string()
        elif isinstance(symbol, EP):
            query = symbol.raw_string
        else:
            query = symbol

        table.row([query])

    if len(table):
        gr_static.node(f"{graph_node}__symbols", str(table))
        gr_static.edge(
            f"{graph_node}__symbols" if not reverse else f"{graph_node}:symbols",
            f"{graph_node}:symbols" if not reverse else f"{graph_node}__symbols",
            style="dashed", constraint="false" if not reverse else "true",
        )


def build_node(graphs: GraphStorage, node: Node, line_limit: Optional[Tuple[int, int]] = None):
    if node.ep is None:
        raise ValueError(f"EP of node {node.id} could not be None")

    table = GVTable()

    if "IN" in node.labels:
        table.row([
            ("<FONT POINT-SIZE=\"10\">" + html.escape(node.labels["IN"]) + "</FONT>", {"port": "in"}),
        ], attrs={
            "border": "1", "fixedsize": "TRUE", "colspan": "3",
        })

    table.row([(node.ep.name, {"port": "ep"})], attrs={"bgcolor": "gray95"})

    if node.tick_type:
        table.cell([node.tick_type])

    if node.symbols:
        table.cell([("[■]", {"port": "symbols"})])

    if node.ep and (node.ep.args or node.ep.kwargs):
        params: List[Tuple[Optional[str], Union[str, NestedQuery]]] = \
            [(None, v) for v in node.ep.args] + list(node.ep.kwargs.values())

        for idx, data in enumerate(params):
            k, v = data
            attrs = {"port": k}
            if idx == len(params) - 1:
                attrs["sides"] = "LRB"
            else:
                attrs["sides"] = "LR"

            if isinstance(v, NestedQuery):
                param_value = v.raw_string
            else:
                param_value = v

            param_value = truncate_param_value(node.ep, k, param_value, line_limit)
            param_value = html.escape(param_value).replace("\t", "&nbsp;" * 4)
            param_lines = param_value.split("\n")

            if k:
                if len(param_lines) == 1:
                    param_lines[0] = f"{html.escape(k)}={param_lines[0]}"
                else:
                    param_lines = [f"{html.escape(k)}="] + param_lines

            if len(param_lines) > 1:
                # Add idents disable default horizontal central align
                # if there are multiline parameter for EP.
                # Align change affects all parameters for EP.
                for i in range(len(param_lines)):
                    if i > 0:
                        param_lines[i] = "&nbsp;" * 2 + param_lines[i]

                attrs.update({"ALIGN": "LEFT", "BALIGN": "LEFT"})

            table.row([param_lines], attrs=attrs)

    if node.params:
        table.row([[f"{html.escape(k)}={html.escape(v)}" for k, v in node.params.items()]])

    if isinstance(node.ep, NestedQuery):
        if not (node.ep.is_local and node.ep.query or graphs.get_query(node.ep.file_path, node.ep.query)):
            table.row([node.ep.to_string()])

    if isinstance(node.ep, IfElseEP):
        table.row([
            ("<FONT POINT-SIZE=\"10\">[IF]</FONT>", {"port": "if"}), ("", {"border": "0"}),
            ("<FONT POINT-SIZE=\"10\">[ELSE]</FONT>", {"port": "else"})
        ], attrs={
            "border": "1", "fixedsize": "TRUE", "colspan": "1",
        })
    elif "OUT" in node.labels:
        table.row([
            ("<FONT POINT-SIZE=\"10\">" + html.escape(node.labels["OUT"]) + "</FONT>", {"port": "out"}),
        ], attrs={
            "border": "1", "fixedsize": "TRUE", "colspan": "3",
        })

    return str(table)


def _parse_time(time_str: str) -> str:
    if time_str:
        try:
            time_str = datetime.strptime(time_str, "%Y%m%d%H%M%S%f").strftime("%Y/%m/%d %H:%M:%S.%f"[:-3])
        except ValueError:
            pass
    else:
        time_str = "--"

    return time_str


def _build_time_expr(table: GVTable, name: str, time_expr: str):
    attrs = {}
    time_expr = html.escape(time_expr)

    if len(time_expr) > 60:
        time_expr = f"<FONT POINT-SIZE=\"10\">{time_expr}</FONT>"
        attrs["cellpadding"] = "4"

    table.row([name], attrs={"bgcolor": "gray95"}).row([time_expr], attrs=attrs)


def _get_nested_query(nested_query: NestedQuery, local_graph: Graph, graphs: GraphStorage) -> Optional[Query]:
    if nested_query.query:
        if nested_query.is_local:
            return local_graph.queries[nested_query.query]
        else:
            return graphs.get_query(nested_query.file_path, nested_query.query)

    return None


def _render_graph(
    gr_root, gr, graphs: GraphStorage, graph_name: str, queries: set, line_limit: Optional[Tuple[int, int]] = None,
):
    graph = graphs[graph_name]

    if not queries or queries == {"*"}:
        queries = set(graph.queries.keys())

    for query_name in queries:
        if query_name not in graph.queries:
            continue

        query = graph.queries[query_name]
        query_id = query.get_id()

        with gr.subgraph(name=query_id, node_attr={"shape": "plaintext"}) as gr_sub:
            gr_sub.attr(label=query_name)

            start_time = _parse_time(query.config.get("START", graph.config.get("START")))
            end_time = _parse_time(query.config.get("END", graph.config.get("END")))

            start_expression = query.config.get("START_EXPRESSION", graph.config.get("START_EXPRESSION"))
            end_expression = query.config.get("END_EXPRESSION", graph.config.get("END_EXPRESSION"))

            tz_data = query.config.get("TZ", graph.config.get("TZ"))
            if not tz_data:
                tz_data = "--"

            table = GVTable().row([
                "START_TIME", "END_TIME", "TZ"
            ], attrs={"bgcolor": "gray95"}).row([
                start_time, end_time, tz_data,
            ])

            if start_expression:
                _build_time_expr(table, "START_EXPRESSION", start_expression)

            if end_expression:
                _build_time_expr(table, "END_EXPRESSION", end_expression)

            table.row([
                ("PARAMETERS", {"port": "params"}),
                ("SYMBOLS", {"port": "symbols"}),
            ], attrs={"bgcolor": "gray95"})

            footer_id = f"{query_id}__footer"
            gr_sub.node(footer_id, str(table), labelloc="c")

            # Put footer to the bottom and, most times, to the center
            for node_id in query.leaves:
                gr_sub.edge(_get_node_unique_id(node_id, query), footer_id, style="invis")

            if query.params:
                gr_sub.node(
                    f"{query_id}__params",
                    str(GVTable().row([[
                        f"{html.escape(k)}" + (f" = {html.escape(v)}" if v else "") for k, v in query.params.items()
                    ]]))
                )

                gr_sub.edge(
                    f"{footer_id}:params", f"{query_id}__params",
                    style="dashed", constraint="true",
                )

            if query.symbols:
                build_symbols(query.symbols, gr, gr_sub, graphs, f"{query_id}__footer", reverse=True)

            for node_id, node in query.nodes.items():
                node_unique_id = _get_node_unique_id(node, query)
                gr_sub.node(node_unique_id, build_node(graphs, node, line_limit), group=query_name)

                for sink in node.sinks:
                    if "OUT" in node.labels:
                        output_port = ":out"
                    else:
                        output_port = ""

                    if isinstance(node.ep, IfElseEP):
                        if sink in node.ep.else_nodes:
                            output_port = ":else"
                        else:
                            output_port = ":if"

                    sink_node = query.nodes[sink]
                    if "IN" in sink_node.labels:
                        sink_port = ":in"
                    else:
                        sink_port = ""

                    gr_sub.edge(
                        f"{node_unique_id}{output_port}", f"{_get_node_unique_id(sink_node, query)}{sink_port}",
                    )

                for param_name, param_value in node.ep.kwargs.items():
                    if isinstance(param_value[1], NestedQuery):
                        nested_cluster = _get_nested_query(param_value[1], graph, graphs)
                        if not nested_cluster:
                            continue

                        gr_root.edge(
                            f"{node_unique_id}:{param_name}",
                            _get_node_unique_id(nested_cluster.roots[0], nested_cluster),
                            lhead=nested_cluster.get_id(),
                            style="dashed", dir="both",
                        )

                if node.symbols:
                    build_symbols(node.symbols, gr, gr_sub, graphs, node_unique_id)

                if isinstance(node.ep, NestedQuery):
                    nested_cluster = _get_nested_query(node.ep, graph, graphs)
                    if not nested_cluster:
                        continue

                    gr_root.edge(
                        node_unique_id,
                        _get_node_unique_id(nested_cluster.roots[0], nested_cluster),
                        lhead=nested_cluster.get_id(),
                        style="dashed", dir="both",
                    )


def render_otq(
    path: Union[str, List[str]],
    image_path: Optional[str] = None,
    output_format: Optional[str] = None,
    load_external_otqs: bool = True,
    view: bool = False,
    line_limit: Optional[Tuple[int, int]] = (10, 30),
    parse_eval_from_params: bool = False,
    debug: bool = False,
) -> str:
    """
    Render queries from .otq files.

    Parameters
    ----------
    path: str, List[str]
        Path to .otq file or list of paths to multiple .otq files.
        Needed to render query could be specified with the next format: `path_to_otq::query_name`
    image_path: str, None
        Path for generated image. If omitted, image will be saved in a temp dir
    output_format: str, None
        `Graphviz` rendering format. Default: `png`.
        If `image_path` contains one of next extensions, `output_format` will be set automatically: `png`, `svg`, `dot`.
    load_external_otqs: bool
        If set to `True` (default) dependencies from external .otq files (not listed in ``path`` param)
        will be loaded automatically.
    view: bool
        Defines should generated image be shown after render.
    line_limit: Tuple[int, int], None
        Limit for maximum number of lines and length of some EP parameters strings.
        First param is limit of lines, second - limit of characters in each line.
        If set to None limit disabled.
        If one of tuple values set to zero the corresponding limit disabled.
    parse_eval_from_params: bool
        Enable parsing and printing `eval` sub-queries from EP parameters.
    debug: bool
        Allow to print stdout or stderr from `Graphviz` render.

    Returns
    -------
    Path to rendered image

    Examples
    --------

    Render single file:

    >>> otp.utils.render_otq("./test.otq")  # doctest: +SKIP

    .. image:: ../../static/testing/images/render_otq_1.png

    Render multiple files:

    >>> otp.utils.render_otq(["./first.otq", "./second.otq"])  # doctest: +SKIP

    .. image:: ../../static/testing/images/render_otq_2.png

    Render specific queries from multiple files:

    >>> otp.utils.render_otq(["./first.otq", "./second.otq::some_query"])  # doctest: +SKIP
    """
    if not isinstance(path, list):
        path = [path]

    path = [_posix_path(p) for p in path]

    graphs = GraphStorage()

    queries_to_render: Dict[str, Set[str]] = defaultdict(set)
    path_files: List[str] = []

    for otq_path in path:
        query_file, query_name = _parse_query_path(otq_path)

        path_files.append(query_file)

        if queries_to_render[query_file] == {"*"}:
            continue

        if query_name:
            queries_to_render[query_file].add(query_name)
        else:
            queries_to_render[query_file] = {"*"}

    otq_files: Deque[str] = deque(path_files)

    while otq_files:
        otq_path = otq_files.popleft()

        graph = read_otq(otq_path, parse_eval_from_params=parse_eval_from_params)

        if not graph:
            continue

        graphs[otq_path] = graph

        for graph_query in graph.queries.values():
            for dep_file, dep_query in graph_query.depends:
                if dep_file is None:
                    dep_file = otq_path

                if dep_file not in graphs and load_external_otqs:
                    otq_files.append(dep_file)

                if queries_to_render[dep_file] == {"*"}:
                    continue

                if (
                    load_external_otqs and dep_file != otq_path or
                    dep_file == otq_path and graph_query.name in queries_to_render[otq_path]
                ):
                    if dep_query:
                        queries_to_render[dep_file].add(dep_query)
                    else:
                        queries_to_render[dep_file] = {"*"}

    if image_path:
        extension = Path(image_path).suffix

        if extension:
            extension = extension[1:]

            if extension == output_format or output_format is None and extension in {"png", "svg", "dot"}:
                image_path = str(Path(image_path).with_suffix(""))

                if output_format is None:
                    output_format = extension

    if not output_format:
        output_format = "png"

    if not image_path:
        image_path = TmpFile().path

    gr = gv.Digraph(format=output_format, filename=image_path, engine="dot")
    gr.attr("graph", compound="true")
    gr.attr("node", shape="plaintext", margin="0")

    idx = 0
    for otq_path, graph in graphs.items():
        if not queries_to_render[otq_path]:
            continue

        with gr.subgraph(name=f"cluster__graph__{idx}", node_attr={"shape": "plaintext"}) as gr_otq:
            gr_otq.attr(label=otq_path)
            gr_otq.attr(margin="16")
            _render_graph(gr, gr_otq, graphs, otq_path, queries_to_render[otq_path], line_limit)

        idx += 1

    return gr.render(view=view, quiet=not debug)



from abc import ABC, abstractmethod
from typing import List, Dict, Union, TYPE_CHECKING, Tuple, Optional, Any
from copy import deepcopy
from functools import wraps
from collections import namedtuple
import pandas as pd

if TYPE_CHECKING:
    from onetick.py.core.source import Source   # hack for annotations

from onetick.py.core.column import _Column
from onetick.py.core.column_operations.base import _Operation, OnetickParameter
from onetick.py.core._source._symbol_param import _SymbolParamColumn
from onetick.py import types as ott
from onetick.py import utils
from onetick.py.otq import otq


def validate(method):
    """wraps schema getter with validations of input columns + src and resulting schema + output column"""

    @wraps(method)
    def inner(obj: '_Aggregation', src: 'Source', name):
        obj.validate_input_columns(src)
        for column in obj.group_by:
            if str(column) not in src.schema or not isinstance(src[str(column)], _Column):
                raise KeyError(f"There is no '{column}' column to group by")
        schema: Dict = method(obj, src=src, name=name)
        if not obj.overwrite_output_field:
            obj.validate_output_name(schema, name)
        return schema

    return inner


def operation_gb(method):
    """wraps aggregation to apply _Operation and remove it after aggregation"""

    @wraps(method)
    def inner(obj: '_Aggregation', src: 'Source', *args, **kwargs) -> 'Source':
        inplace = kwargs.get('inplace')
        res = src if inplace else src.copy()
        src_schema = src.schema

        gb_copy = obj.group_by.copy()
        obj.group_by = []
        for i, gb in enumerate(gb_copy):
            if isinstance(gb, _Operation) and not isinstance(gb, _Column):
                name = f'GROUP_{i}'
                if name in src_schema:
                    raise AttributeError(f"'{name}' column name is reserved for group by Operation "
                                         f"but it exists in current schema")
                res[name] = gb
                obj.group_by.append(res[name])
            else:
                obj.group_by.append(gb)
        res = method(obj, res, *args, **kwargs)

        obj.group_by = gb_copy
        return res
    return inner


def operation_replacer(method):
    """
    PY-378
    Decorator allows working with aggregation's columns specified as operations.
    """

    @wraps(method)
    def inner(obj: '_Aggregation', src: 'Source', *args, **kwargs) -> 'Source':
        inplace = kwargs.get('inplace')
        res = src if inplace else src.copy()
        tmp_columns = {}

        aggrs = getattr(obj, 'aggrs', None)
        if aggrs:
            aggs = aggrs.values()
        else:
            name = args[0] if args else kwargs.get('name')
            # pylint: disable-next=unidiomatic-typecheck
            if type(obj.column_name) is _Operation and name is None:
                raise ValueError('Output field name must be specified when aggregating operation')
            aggs = [obj]

        # Add operation from each aggregation object to source `res` as column
        # and replace *column_name* property in each aggregation object with this column's name.
        for i, agg in enumerate(aggs):
            # pylint: disable-next=unidiomatic-typecheck
            if type(agg.column_name) is _Operation:
                tmp_name = f'__TMP_AGG_COLUMN_{i}__'
                res[tmp_name] = agg.column_name
                tmp_columns[tmp_name] = (agg, agg.column_name)
                agg.column_name = tmp_name

        res = method(obj, res, *args, **kwargs)

        if tmp_columns:
            # Rollback all aggregation objects and source `res`.
            # Delete all temporary columns and change property *column_name* back in aggregations.
            to_drop = list(set(tmp_columns).intersection(res.schema))
            if to_drop:
                res.drop(to_drop, inplace=True)
            for agg, column_name in tmp_columns.values():
                agg.column_name = column_name

        return res
    return inner


def output_column_overwriter(method):
    """
    Allows outputting aggregation to existing field.
    In this case temporary renaming existing field.
    """
    @wraps(method)
    def inner(obj: '_Aggregation', src: 'Source', *args, **kwargs) -> 'Source':
        column_name = obj.column_name
        name = args[0] if args else kwargs.get('name')
        name = name or column_name

        if not obj.overwrite_output_field or not name or name not in src.schema:
            return method(obj, src, *args, **kwargs)

        inplace = kwargs.get('inplace')
        res = src if inplace else src.copy()

        # rename existing field to the temporary name
        tmp_name = f'__TMP_AGG_COLUMN_{name}__'
        res[tmp_name] = res[name]
        res.drop(name, inplace=True)
        # aggregating renamed field
        kwargs['name'] = name
        obj.column_name = tmp_name

        res = method(obj, res, *args, **kwargs)

        # removing temporary field
        if tmp_name in res.schema:
            res.drop(tmp_name, inplace=True)
        obj.column_name = column_name

        return res
    return inner


def get_seconds_from_time_offset(time_offset):
    if not isinstance(time_offset, ott.OTPBaseTimeOffset):
        raise ValueError('Only DatePart objects can be passed in this function')

    return int(pd.Timedelta(time_offset).total_seconds())


def get_bucket_interval_from_datepart(bucket_interval):
    if not isinstance(bucket_interval, ott.OTPBaseTimeOffset):
        raise ValueError('Only DatePart objects can be passed in this function')

    if isinstance(bucket_interval, ott.ExpressionDefinedTimeOffset):
        raise ValueError(f"Operation as DatePart isn't allowed: {str(bucket_interval.n)}")

    # bucket_interval also could be one of these:
    # otp.Milli, otp.Second, otp.Minute, otp.Hour, otp.Day, otp.Month
    # bucket_interval will be converted and corresponding bucket_units value will be set

    offset, datepart = bucket_interval.get_offset()
    if datepart not in {'millisecond', 'second', 'minute', 'hour', 'day', 'month'}:
        raise ValueError(f"Unsupported DatePart passed to bucket_interval: {datepart}")

    if offset < 0:
        raise ValueError(
            f"Negative DateParts aren't allowed for bucket_interval: {offset} ({datepart})"
        )

    if datepart in {'millisecond', 'minute', 'hour'}:
        # bucket_units could be only seconds, days, months or ticks
        # so other DateParts are converted to seconds
        if datepart == 'millisecond':
            offset, datepart = offset / 1000, 'second'
        else:
            offset, datepart = ott.Second(get_seconds_from_time_offset(bucket_interval)).get_offset()

    return offset, f"{datepart}s"  # type: ignore[union-attr]


class _Aggregation(ABC):

    @property
    @abstractmethod
    def NAME(self) -> str:
        pass

    @property
    @abstractmethod
    def EP(self) -> otq.EpBase:
        pass

    DEFAULT_OUTPUT_NAME = 'VALUE'

    FIELDS_MAPPING = {
        "column_name": "INPUT_FIELD_NAME",
        "running": "IS_RUNNING_AGGR",
        "all_fields": "ALL_FIELDS_FOR_SLIDING",
        "bucket_interval": "BUCKET_INTERVAL",
        "bucket_time": "BUCKET_TIME",
        "bucket_units": "BUCKET_INTERVAL_UNITS",
        "bucket_end_condition": "BUCKET_END_CRITERIA",
        "end_condition_per_group": "BUCKET_END_PER_GROUP",
        "boundary_tick_bucket": "BOUNDARY_TICK_BUCKET",
        "group_by": "GROUP_BY",
        "groups_to_display": "GROUPS_TO_DISPLAY",
    }
    FIELDS_DEFAULT = {
        "running": False,
        "all_fields": False,
        "bucket_interval": 0,
        "bucket_time": "BUCKET_END",
        "bucket_units": "seconds",
        "bucket_end_condition": None,
        "end_condition_per_group": False,
        "boundary_tick_bucket": "new",
        "group_by": [],
        "groups_to_display": "all",
    }

    FIELDS_TO_SKIP: List = []   # attr listed here won't be used in self.__str__

    output_field_type: Optional[type] = None  # None will force to use type of input column
    require_type: Optional[Tuple[type, ...]] = None
    _validations_to_skip: List = []

    def __init__(self,
                 column: Union[str, _Column, _Operation],
                 running: bool = False,
                 all_fields: Union[bool, str] = False,
                 bucket_interval: Union[int, ott.OTPBaseTimeOffset] = 0,
                 bucket_time: str = "end",
                 bucket_units: Union[str, None] = None,
                 bucket_end_condition: Optional[_Operation] = None,
                 end_condition_per_group: bool = False,
                 boundary_tick_bucket: str = "new",
                 group_by: Optional[Union[List, str, _Operation]] = None,
                 groups_to_display: str = "all",
                 overwrite_output_field: bool = False):
        """
        Abstract method that implements common logic for aggregations
        """
        if isinstance(column, list):
            column = ','.join(map(str, column))

        column_name: Union[str, _Operation] = str(column)

        if column_name == "Time":
            # TODO: need to understand how to better work with alias
            column_name = "TIMESTAMP"

        # pylint: disable-next=unidiomatic-typecheck
        if type(column) is _Operation:
            column_name = column

        if isinstance(bucket_interval, float):
            if bucket_units is not None and bucket_units != 'seconds':
                raise ValueError('Float values for bucket_interval are only supported for seconds.')
            if bucket_interval < 0.001:
                raise ValueError('Float values for bucket_interval less than 0.001 are not supported.')

        if isinstance(bucket_interval, ott.OTPBaseTimeOffset):
            bucket_interval, bucket_units = get_bucket_interval_from_datepart(bucket_interval)

        if isinstance(all_fields, str) and all_fields == "when_ticks_exit_window":
            if not running:
                raise ValueError("`all_fields` can't be set to 'when_ticks_exit_window' when `running=False`")

            if not bucket_interval:
                raise ValueError(
                    "`all_fields` can't be set to 'when_ticks_exit_window' when `bucket_interval` is zero`"
                )

            all_fields = all_fields.upper()

        self.column_name = column_name
        self.running = running
        self.all_fields = all_fields
        self.bucket_time = bucket_time

        if isinstance(bucket_interval, _Operation):
            if bucket_interval.dtype is bool:
                if bucket_end_condition is not None:
                    raise ValueError(
                        "Bucket end condition passed on both `bucket_interval` and `bucket_end_condition` parameters"
                    )

                bucket_end_condition = bucket_interval
                bucket_interval = 0
            elif isinstance(bucket_interval, OnetickParameter) and bucket_interval.dtype is int:
                bucket_interval = str(bucket_interval)
            elif isinstance(bucket_interval, _SymbolParamColumn) and bucket_interval.dtype is int:
                bucket_interval = str(bucket_interval.expr)
            else:
                raise ValueError("Bucket interval can only be boolean otp.Operation or integer otp.param")

        self.bucket_interval = bucket_interval

        if bucket_end_condition is None:
            self.bucket_end_condition = None  # type: ignore
        else:
            self.bucket_end_condition = str(bucket_end_condition)

        self.bucket_units = bucket_units
        if self.bucket_units is None:
            if self.bucket_end_condition:
                # allow omitting bucket_units if bucket_end_condition is set
                self.bucket_units = 'flexible'
            else:
                # default value
                self.bucket_units = 'seconds'

        self.end_condition_per_group = end_condition_per_group
        self.boundary_tick_bucket = boundary_tick_bucket
        self.large_ints = False
        if isinstance(group_by, (_Operation, str)):
            group_by = [group_by]
        self.group_by = group_by or []
        self.groups_to_display = groups_to_display
        self.overwrite_output_field = overwrite_output_field

        self._param_validation()
        self.bucket_time = f'BUCKET_{self.bucket_time.upper()}'

    @staticmethod
    def _attr2str(value) -> str:
        if isinstance(value, bool):
            return 'true' if value else 'false'
        if isinstance(value, list):
            return ','.join(value)
        return str(value)

    @property
    def ep_params(self) -> Dict:
        """prepare params for self.__str__ and otq.EpBase"""
        params = {}

        for field, ep_param in self.FIELDS_MAPPING.items():
            if field in self.FIELDS_TO_SKIP:
                continue

            default_value = self.FIELDS_DEFAULT.get(field)
            if getattr(self, field) != default_value:
                if field == 'group_by':
                    params[ep_param] = ",".join(list(map(str, self.group_by)))
                else:
                    params[ep_param] = getattr(self, field)
        return params

    def __str__(self):
        params = [f'{k}={self._attr2str(v)}' for k, v in self.ep_params.items()]
        return self.NAME + "(" + ",".join(params) + ")"

    def to_ep(self, name: Optional[str]) -> otq.EpBase:
        params = dict((k.lower(), v) for k, v in self.ep_params.items())
        if 'output_field_name' not in self.FIELDS_TO_SKIP:
            params['output_field_name'] = name
        return self.EP(**params)

    @validate
    def _get_common_schema(self, src: 'Source', name: str) -> Dict:
        """return data schema without output fields (this fields should be added further)"""
        schema = {}
        for column in self.group_by:
            schema[str(column)] = src.schema[str(column)]
        if self.all_fields:
            schema.update(src.schema)
        return schema

    def _modify_source(self, res: 'Source', **kwargs):
        """
        Modify resulting source inplace before sinking to aggregation.
        Can be overriden if needed.
        """
        pass

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> Dict:
        if not name or name in src.__class__.meta_fields:
            return {}
        return {
            name: self.output_field_type or src.schema[self.column_name]
        }

    @operation_gb
    @operation_replacer
    @output_column_overwriter
    def apply(self, src: 'Source', name: Optional[str] = None, inplace: bool = False) -> 'Source':
        """
        Applies aggregation to Source and sets proper schema

        Parameters
        ----------
        src: Source
            Source to apply aggregation
        name: str, optional
            Name of output column. If not specified, will be used self.column_name
        inplace: bool
            Modify passed ``src`` object or return modified copy.
        """
        if inplace:
            res = src
            src = src.copy()
        else:
            res = src.copy()
        out_name = name or self.column_name
        schema = self._get_common_schema(src, out_name)
        # it's important to validate input schema before sinking
        self._modify_source(res)
        res.sink(self.to_ep(name=str(out_name)))
        schema.update(self._get_output_schema(src, str(out_name)))
        res.schema.set(**schema)

        if not self.all_fields:
            # in this case we propagate only resulting fields, that stored in res.schema (flexible schema case)
            res._add_table(strict=True)
        else:
            # adding table to convert types in schema, e.g. float to int
            res._add_table(strict=False)
        return res

    def validate_input_columns(self, src: 'Source'):
        """checks that columns used in aggregation presented in Source"""
        if self.column_name not in src.schema:
            raise TypeError(f"Aggregation `{self.NAME}` uses column `{self.column_name}` as input, which doesn't exist")
        if not self.require_type:
            return
        dtype = src.schema[self.column_name]
        base_dtype = ott.get_base_type(dtype)
        for t in self.require_type:
            # more generic types can be specified in self.require_type too
            if dtype is t or base_dtype is t:
                return
        raise TypeError(f"Aggregation `{self.NAME}` require {self.require_type} types, got {dtype}")

    @staticmethod
    def validate_output_name(schema: Dict, name: Union[List, str]):
        """checks that aggregation won't output columns with same names"""
        if not isinstance(name, list):
            name = [name]

        same_fields = []
        for n in name:
            if n in schema:
                if '__long_nsec_' in n:
                    same_fields.append(n.replace('__long_nsec_', ''))   # hack for large ints
                else:
                    same_fields.append(n)
        if same_fields:
            raise ValueError("You try to propagate all fields and put result into already existing fields: "
                             f"'{', '.join(same_fields)}' ")

    def _param_validation(self):
        """validate __init__ parameters"""
        if self.running and self.bucket_time == "start":
            raise ValueError("It is not allowed to set up running=True and bucket_time='start'")
        if self.bucket_units == "flexible" and self.bucket_end_condition is None:
            raise ValueError("bucket_units is set to 'flexible' but bucket_end_condition is not specified. "
                             "Please specify bucket_end_condition.")
        if self.bucket_units != "flexible" and self.bucket_end_condition is not None:
            raise ValueError("bucket_end_condition can be used only with 'flexible' bucket_units. "
                             "Please set bucket_units to 'flexible'.")

        if self.bucket_time not in ['start', 'end']:
            raise ValueError(f"'bucket_time' might be either 'start' or 'end', but passed '{self.bucket_time}'")

        valid_units = ("seconds", "ticks", "days", "months", "flexible")
        if self.bucket_units not in valid_units:
            raise ValueError("'bucket_units' can be one of the following: "
                             f"'{', '.join(valid_units)}'; however, '{self.bucket_units}' was passed")

        valid_boundary = {"new", "previous"}
        if self.boundary_tick_bucket not in valid_boundary:
            message = "'boundary_tick_bucket' can be one of the following: {}; however, {} was passed"
            raise ValueError(message.format(', '.join(list(valid_boundary)), self.boundary_tick_bucket))

        for column in self.group_by:
            if not isinstance(column, _Operation) and not isinstance(column, str):
                raise TypeError(f"Unsupported type '{column}' of a column to group by")

        if self.groups_to_display not in ('all', 'event_in_last_bucket'):
            raise ValueError("Parameter 'groups_to_display' can only be set to 'all' or 'event_in_last_bucket':"
                             f" got '{self.groups_to_display}'")

        if self.all_fields and not self.running and 'running_all_fields' not in self._validations_to_skip:
            raise ValueError("It is not allowed set all_fields to True for not running aggregation")

        if not self.running and self.overwrite_output_field:
            raise ValueError("Parameter 'overwrite_output_field' can only be used with running aggregations")

    @property
    def is_multi_column_aggregation(self):
        return isinstance(self, _MultiColumnAggregation)

    @property
    def is_all_columns_aggregation(self):
        return isinstance(self, _AllColumnsAggregation)


class _AggregationTSType(_Aggregation):

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['time_series_type'] = 'TIME_SERIES_TYPE'
    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['time_series_type'] = 'event_ts'

    def __init__(self, column, time_series_type: str = "event_ts", *args, **kwargs):
        """
        Abstract class that implements common logic for aggregations with ability to select time series type
        inherited from _Aggregation

        Parameters
        ----------
        column: see _Aggregation
        time_series_type: "event_ts" or "state_ts", default="event_ts"
            "state_ts":
                if there is a tick in bucket with timestamp = bucket start:
                    only ticks in bucket used for calculation max value
                else:
                    latest tick from previous bucket included in current bucket
            "event_ts": only ticks from current bucket used for calculations
        args: see _Aggregation
        kwargs: see _Aggregation
        """
        if time_series_type not in ["event_ts", "state_ts"]:
            raise ValueError('time_series_type argument must be "event_ts" or "state_ts"')
        self.time_series_type = time_series_type
        super().__init__(column, *args, **kwargs)


class _AggregationTSSelection(_Aggregation):

    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['selection'] = 'SELECTION'
    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['selection'] = 'first'

    def __init__(self, column, selection: str = "first", *args, **kwargs):
        if selection not in ["first", "last"]:
            raise ValueError(f'{self.__class__.__name__} selection argument must be "first" or "last"')
        self.selection = selection
        super().__init__(column, *args, **kwargs)


class _FloatAggregation(_Aggregation):

    require_type = (int, float, ott._inf)

    """
    Aggregation that expect int or float as input
    """


class _KeepTs(_Aggregation):

    def __init__(self, *args, keep_timestamp=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.keep_timestamp = keep_timestamp

    @validate     # type: ignore
    def _get_common_schema(self, src: 'Source', *args, **kwargs) -> Dict:
        schema = src.schema.copy()
        schema['TICK_TIME'] = ott.nsectime
        return schema

    def apply(self, src: 'Source', *args, **kwargs) -> 'Source':
        res = super().apply(src=src, *args, **kwargs)
        if self.keep_timestamp:
            # TICK_TIME can be empty if it's a tick from default_tick aggregation parameter
            res['TICK_TIME'] = res.if_else(res['TICK_TIME'], res['TICK_TIME'], res['TIMESTAMP'])
            res['TIMESTAMP'] = res['TICK_TIME']
            res.drop('TICK_TIME', inplace=True)
        return res


class _ExpectLargeInts(_Aggregation):
    FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
    FIELDS_MAPPING['large_ints'] = 'EXPECT_LARGE_INTS'
    FIELDS_MAPPING['null_int_val'] = 'NULL_INT_VAL'
    FIELDS_DEFAULT = deepcopy(_Aggregation.FIELDS_DEFAULT)
    FIELDS_DEFAULT['large_ints'] = False
    FIELDS_DEFAULT['null_int_val'] = 0

    def __init__(self, *args, large_ints=False, null_int_val=0, **kwargs):
        super().__init__(*args, **kwargs)
        if large_ints not in {True, False, utils.adaptive}:
            raise ValueError(f"Wrong value for {self.__class__.__name__} aggregation"
                             f" 'large_ints' parameter: {large_ints}")
        if large_ints is utils.adaptive:
            large_ints = 'IF_INPUT_VAL_IS_LONG_INTEGER'

        if null_int_val and not large_ints:
            raise ValueError(
                f"Wrong value for {self.__class__.__name__} aggregation:"
                f" 'null_int_val' parameter is set, however 'large_ints' is `False`"
            )

        self.large_ints = large_ints
        self.null_int_val = null_int_val

    def apply(self, src: 'Source', name: Optional[str] = None) -> 'Source':
        out_name = name or self.column_name
        res, col, convert_back = self._ts_to_long(src, str(out_name))
        res = super().apply(res, col.tmp_out_column)
        if not convert_back:
            return res
        return self._long_to_ts(res, col)

    def _ts_to_long(self, src: 'Source', name: str) -> Tuple['Source', Any, bool]:
        agg_columns = namedtuple('agg_columns', ('in_column', 'tmp_in_column', 'tmp_out_column', 'out_column'))
        if src.schema[self.column_name] != ott.nsectime:
            return src, agg_columns(self.column_name, self.column_name, name, name), False
        self.large_ints = True
        res = src.copy()
        col = agg_columns(self.column_name, f'__long_nsec_{self.column_name}',
                          f'__long_nsec_{name}', name)
        res[col.tmp_in_column] = res[col.in_column].apply(int)
        self.column_name = col.tmp_in_column
        return res, col, True

    def _long_to_ts(self, src: 'Source', col) -> 'Source':
        res = src.copy()
        res[col.out_column] = res[col.tmp_out_column].astype(ott.nsectime)
        to_drop = []
        for c in [col.tmp_out_column, col.tmp_in_column]:
            if c in res.schema:
                to_drop.append(c)
        if to_drop:
            res.drop(to_drop, inplace=True)
        self.column_name = col.in_column
        return res


class _MultiColumnAggregation:
    """
    Helper class for identifying multi-column aggregations.
    """
    pass


class _AllColumnsAggregation(_MultiColumnAggregation):
    """
    Helper class for identifying aggregations, which returns all fields from original ticks.
    """
    pass



from .helpers import (
    get_symbol_list_from_df,
    JSONEncoder,
    json_dumps,
    query_properties_to_dict,
    query_properties_from_dict,
    symbol_date_to_str,
)
from .temp import (
    File,
    PermanentFile,
    TmpFile,
    TmpDir,
    GeneratedDir,
    TMP_CONFIGS_DIR,
    ONE_TICK_TMP_DIR,
)
from .default import (
    get_local_number_of_cores,
    default_license_dir,
    default_license_file,
    default_day_boundary_tz,
)
from .acl import (
    if_db_in_acl,
    add_user_to_acl,
    remove_user_from_acl,
    if_user_in_acl,
    tmp_acl,
)
from .config import (
    reload_config,
    modify_config_param,
    get_config_param,
    is_param_in_config,
    tmp_config,
)
from .locator import (
    get_dbs_locations_from_locator,
    tmp_locator,
    empty_locator,
)
from .script import (
    write_config_for_tick_server_run,
    create_file_to_run_config,
    omd_dist_path,
)
from .query import (
    abspath_to_query_by_otq_path,
    abspath_to_query_by_name,
    query_to_path_and_name,
)
from .types import (
    get_type_that_includes,
    adaptive,
    adaptive_to_default,
    default,
    range,
)
from .tz import (
    get_tzfile_by_name,
    get_timezone_from_datetime,
    convert_timezone,
)
from .file import (
    FileBuffer,
    file,
)
from .render import render_otq
from . import perf



import os
import traceback
from typing import List

from .configuration import config


# dictionary to save traceback lines
_TRACE_LINES = {}
# dictionary to save traceback tuples
_TRACE_TUPLES = {}


def _get_id_with_traceback(tb_list: List[str]) -> str:
    """
    Save traceback in some memory-efficient way
    and return unique id of the traceback.
    """
    trace_list = []
    for line in tb_list:
        if line not in _TRACE_LINES:
            # using the same string objects for different tracebacks
            _TRACE_LINES[line] = line
        trace_list.append(_TRACE_LINES[line])

    trace_tuple = tuple(trace_list)
    trace_hash = str(hash(trace_tuple))
    if trace_hash not in _TRACE_TUPLES:
        # using the same tuple objects for different tracebacks
        _TRACE_TUPLES[trace_hash] = trace_tuple
    return trace_hash


def _get_traceback_with_id(trace_hash: str) -> str:
    """
    Get our custom saved traceback from dictionary by hash.
    """
    return ''.join(_TRACE_TUPLES[trace_hash])


def _modify_stack_info_in_onetick_query():
    """
    Change stack_info parameter in all OneTick's event processors.
    Save full traceback instead of filename + line number.
    """

    def modify_init(cls):
        old_init = cls.__init__

        def new_init(self, *args, **kwargs):
            old_init(self, *args, **kwargs)
            if config.show_stack_info and hasattr(self, 'stack_info'):
                self.stack_info = _get_id_with_traceback(traceback.format_stack()[:-1])

        cls.__init__ = new_init

    from onetick.py.otq import otq
    eps_classes = (
        cls for cls in otq.__dict__.values()
        if isinstance(cls, type) and issubclass(cls, otq.EpBase) and cls is not otq.EpBase
    )
    for ep_cls in eps_classes:
        modify_init(ep_cls)


def _add_stack_info_to_exception(exc):
    """
    Find stack_info parameter in OneTick exception message.
    Get stack trace and append it to the passed exception.
    """
    if not config.show_stack_info:
        return exc

    _, _, location_details = str(exc).partition('Problem location details:')
    stack_info_uuid = None
    for block in location_details.strip().split(','):
        name, _, value = block.partition('=')
        if name == 'stack_info':
            stack_info_uuid = value
            break

    if not stack_info_uuid:
        return exc

    stack_info = _get_traceback_with_id(stack_info_uuid)

    if exc.args:
        exc.args = (exc.args[0] + os.linesep + stack_info, *exc.args[1:])
    return exc



import os
import zlib

from pathlib import Path
from typing import Union


class FileBuffer:
    '''
    Class holds the file content with goal to delivery
    it to the execution side in case of remote executions.

    The basic implementation reads file content to a property
    that allows to transfer file content as pickled object
    to the server side since the pickling stores all class property
    values.
    '''

    def __init__(self, path: Union[str, os.PathLike]):
        content = Path(path).read_text()
        self._content = zlib.compress(content.encode('utf-8'), level=9)

    def get(self):
        ''' Returns file content '''
        return zlib.decompress(self._content).decode('utf-8')


def file(path: Union[str, os.PathLike]) -> FileBuffer:
    '''
    Helps to build a file buffer that could be used to
    delivery on the remote site to be processed there.
    For example it could be passed as input to the :class:`CSV <onetick.py.CSV>`

    See Also
    --------
        :class:`CSV <onetick.py.CSV>`
    '''
    return FileBuffer(path)



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from onetick.py.core.source import Source   # hack for annotations

import onetick.py as otp
from onetick.py.otq import otq
from onetick.py.core.column import _Column
from ._base import _Aggregation, _MultiColumnAggregation


class Generic(_Aggregation, _MultiColumnAggregation):
    NAME = 'GENERIC_AGGREGATION'
    EP = otq.GenericAggregation

    FIELDS_TO_SKIP = ['column_name', 'all_fields', 'output_field_name']
    FIELDS_MAPPING = dict(_Aggregation.FIELDS_MAPPING, **{
        'query_name': 'QUERY_NAME',
        'bucket_delimiter': 'BUCKET_DELIMITERS',
    })
    FIELDS_DEFAULT = dict(_Aggregation.FIELDS_DEFAULT, **{
        'bucket_delimiter': None,
    })
    _validations_to_skip = ['running_all_fields']

    def __init__(self,
                 query_fun,
                 bucket_delimiter: bool = False,
                 **kwargs):
        self._query: Optional['Source'] = None
        self._query_fun = query_fun
        self._query_params: Optional[dict] = None
        self.bucket_delimiter = 'D' if bucket_delimiter else None

        # init variables which will be set later
        self.query_name: Optional[str] = None
        self._query_schema: Optional[dict] = None

        if 'all_fields' in kwargs:
            raise ValueError("Parameter 'all_fields' for generic aggregation is meaningless. "
                             "Aggregated source will have all fields returned by 'query_fun'.")
        # we don't want to set hard limit on the output of order book aggregations
        kwargs['all_fields'] = True
        super().__init__(column=_Column('TIMESTAMP'), **kwargs)

    def _set_query_params(self, **kwargs):
        self._query_params = kwargs

    def apply(self, src: 'Source', name: Optional[str] = None, **kwargs) -> 'Source':
        """Applies generic aggregation to Source and sets proper schema

        Parameters
        ----------
        src: Source
            Source to apply aggregation
        name: str, optional
            Name of output column. If not specified, will be used self.column_name
        kwargs: dict
            Parameters to be passed to `query_fun()` when creating aggregation query
        """
        self._set_query_params(**kwargs)
        return super().apply(src, name=name)

    def _make_query_object(self, schema):
        query_params = self._query_params if self._query_params else {}
        query = otp.DataSource(symbols='LOCAL::', tick_type='ANY', schema_policy='manual', schema=schema)
        query = self._query_fun(query, **query_params)
        return query

    def _detect_query_fun_schema(self, res):
        # this will be translated to passthrough with symbol and tick type set
        if self._query is None:
            self._query = self._make_query_object(res.schema)

        return self._query.schema.copy()

    def _modify_source(self, res: 'Source', **kwargs):
        query_schema = self._detect_query_fun_schema(res)

        # schema will be used when validating output
        if self._query_schema is None:
            self._query_schema = query_schema

        if self._query is None:
            raise RuntimeError('Attempted to use `self._query` before initialization')

        # query_name will be used to create ep
        query_name = self._query._store_in_tmp_otq(
            res._tmp_otq, operation_suffix='generic_aggregation', add_passthrough=False,
        )
        self.query_name = f'THIS::{query_name}'

    def _get_common_schema(self, src, name):
        super()._get_common_schema(src, name)
        return {
            column: src.schema[column]
            for column in map(str, self.group_by)
        }

    def _get_output_schema(self, src, name=None):
        schema = self._query_schema.copy()
        if self.bucket_delimiter:
            schema['DELIMITER'] = str
        return schema



from onetick.py import configuration
from onetick.py.otq import otq


class MultiOutputSource:
    """
    Construct a source object with multiple outputs
    from several connected :py:class:`~onetick.py.Source` objects.

    This object can be saved to disk as a graph using :py:meth:`~onetick.py.Source.to_otq` method,
    or passed to :py:func:`onetick.py.run` function.

    If it's passed to :py:func:`onetick.py.run`,
    then returned results for different outputs will be available as a dictionary.

    Parameters
    ----------
    outputs : dict
        Dictionary which keys are names of the output sources, and values are output sources themselves.
        All the passed sources should be connected.

    Examples
    --------

    Results for individual outputs can be accessed by output names

    >>> # OTdirective: skip-snippet:;
    >>> root = otp.Tick(A=1)
    >>> branch_1 = root.copy()
    >>> branch_2 = root.copy()
    >>> branch_3 = root.copy()
    >>> branch_1['B'] = 1
    >>> branch_2['B'] = 2
    >>> branch_3['B'] = 3
    >>> src = otp.MultiOutputSource(dict(BRANCH1=branch_1, BRANCH2=branch_2, BRANCH3=branch_3))
    >>> res = otp.run(src)
    >>> sorted(list(res.keys()))
    ['BRANCH1', 'BRANCH2', 'BRANCH3']
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH1'][['A', 'B']]
       A  B
    0  1  1
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH2'][['A', 'B']]
       A  B
    0  1  2
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH3'][['A', 'B']]
       A  B
    0  1  3

    node_name parameter of the otp.run() method can be used to select outputs

    >>> # OTdirective: skip-snippet:;
    >>> src = otp.MultiOutputSource(dict(BRANCH1=branch_1, BRANCH2=branch_2, BRANCH3=branch_3))
    >>> res = otp.run(src, node_name=['BRANCH2', 'BRANCH3'])
    >>> sorted(list(res.keys()))
    ['BRANCH2', 'BRANCH3']
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH2'][['A', 'B']]
       A  B
    0  1  2
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH3'][['A', 'B']]
       A  B
    0  1  3

    If only one output is selected, then it's returned directly and not in a dictionary

    >>> # OTdirective: skip-snippet:;
    >>> src = otp.MultiOutputSource(dict(BRANCH1=branch_1, BRANCH2=branch_2, BRANCH3=branch_3))
    >>> res = otp.run(src, node_name='BRANCH2')
    >>> res[['A', 'B']]
       A  B
    0  1  2

    A dictionary with sources can also be passed to otp.run directly,
    and MultiOutputSource object will be constructed internally

    >>> res = otp.run(dict(BRANCH1=branch_1, BRANCH2=branch_2))
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH1'][['A', 'B']]
       A  B
    0  1  1
    >>> # OTdirective: skip-snippet:;
    >>> res['BRANCH2'][['A', 'B']]
       A  B
    0  1  2
    """

    def __init__(self, outputs, main_branch_name=None):

        # 1. Checking that outputs have a common part:
        # we create a set of keys for all outputs and see if all sets are connected;
        # two sets are connected if they have any key in common

        if len(outputs) <= 1:
            raise ValueError('At least two branches should be passed to a MultiOutputSource object')

        def get_history_key_set(hist):
            keys = set()
            for rule in hist._rules:
                if "key" in rule.key_params:
                    keys.add(rule.key)
            return keys

        source_key_sets = []
        for source in outputs.values():
            source_key_sets.append(get_history_key_set(source.node()._hist))

        while len(source_key_sets) > 1:
            # we take first set from the list and add to it all the other sets that have common keys with it
            # we continue to do this until first set is the only set in the list or until it has no common keys
            # with other sets in the list
            new_key_sets = []
            first_key_set = source_key_sets[0]
            new_key_sets.append(first_key_set)
            for s in source_key_sets[1:]:
                if first_key_set.isdisjoint(s):
                    # no common keys
                    new_key_sets.append(s)
                else:
                    # there are common keys
                    first_key_set = first_key_set | s
            # checking if first_key_set had common keys with at least some other set
            if len(source_key_sets) == len(new_key_sets):
                raise ValueError("Cannot construct a MultiOutputSource object from outputs that are not connected!")
            # moving first_key_set to the end; maybe it will make things work faster
            new_key_sets = new_key_sets[1:] + [first_key_set]
            source_key_sets = new_key_sets

        # 2, 3. Assigning node names and selecting main branch
        self.__main_branch_name = None
        self.__main_branch = None
        self.__side_branches = {}
        for node_name, source in outputs.items():
            source = source.copy()
            # this is necessary to create different branches if a source is a branching point
            source.sink(otq.Passthrough())
            source.node().node_name(node_name)
            if self.__main_branch_name is None and (main_branch_name is None or main_branch_name == node_name):
                self.__main_branch_name = node_name
                self.__main_branch = source
            else:
                self.__side_branches[node_name] = source
        if self.__main_branch_name is None:
            raise ValueError(f'Branch name "{main_branch_name}" not found among passed outputs!')

        # 4, 5. Apply other branches to the main branch and copy dicts
        self.__main_branch._apply_side_branches(self.__side_branches.values())

    def _all_node_names(self):
        return [self.__main_branch_name] + list(self.__side_branches.keys())

    def _side_branch_list(self):
        return list(self.__side_branches.values())

    def _prepare_for_execution(self, symbols=None, start=None, end=None, start_time_expression=None,
                               end_time_expression=None, timezone=None,
                               has_output=None,  # NOSONAR
                               running_query_flag=None, require_dict=False, node_name=None,
                               symbol_date=None):

        has_output = False  # to avoid sinking PASSTHROUGH to the main branch
        if node_name is None:  # if user passed a node name, we shouldn't overwrite it
            node_name = self._all_node_names()
        return self.__main_branch._prepare_for_execution(
            symbols=symbols, start=start, end=end, start_time_expression=start_time_expression,
            end_time_expression=end_time_expression, timezone=timezone, has_output=has_output,
            running_query_flag=running_query_flag, require_dict=require_dict,
            node_name=node_name,
            symbol_date=symbol_date,
        )

    def to_otq(self, file_name=None, file_suffix=None, query_name=None, symbols=None, start=None, end=None,
               timezone=None):
        """
        Constructs an onetick query graph and saves it to disk

        See Also
        --------
        :py:meth:`onetick.py.Source.to_otq`
        """
        if timezone is None:
            timezone = configuration.config.tz
        return self.__main_branch.to_otq(file_name=file_name,
                                         file_suffix=file_suffix,
                                         query_name=query_name,
                                         symbols=symbols,
                                         start=start,
                                         end=end,
                                         timezone=timezone,
                                         add_passthrough=False)



from typing import List
from sphinx.ext.napoleon.docstring import NumpyDocstring
from sphinx.locale import get_translation
from onetick.doc_utilities.snippets import parse_string


class OTNumpyDocstring(NumpyDocstring):

    def _parse_examples_section(self, section: str) -> List[str]:

        """Applies OT directives for all lines"""

        # combining super()._parse_examples_section and super()._parse_generic_section
        # due to super()._parse_examples_section just prepare parameters and
        # super()._parse_generic_section get lines and modify them.
        # but we need to add logic on lines but before super()._parse_generic_section modification
        labels = {
            'example': get_translation('sphinx')('Example'),
            'examples': get_translation('sphinx')('Examples'),
        }
        use_admonition = self._config.napoleon_use_admonition_for_examples
        section = labels.get(section.lower(), section)

        lines = self._strip_empty(self._consume_to_next_section())
        lines = self._dedent(lines)

        # OneTick custom logic start
        doc = parse_string('\n'.join(lines), caller='doc')
        lines = doc.split('\n')
        # OneTick custom logic end

        if use_admonition:
            header = '.. admonition:: %s' % section     # noqa
            lines = self._indent(lines, 3)
        else:
            header = '.. rubric:: %s' % section     # noqa
        if lines:
            return [header, ''] + lines + ['']
        else:
            return [header, '']
